/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: nickserv.c,v 1.30 2003/03/21 05:17:17 demond Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "alloc.h"
#include "chanserv.h"
#include "client.h"
#include "conf.h"
#include "config.h"
#include "channel.h"
#include "dcc.h"
#include "err.h"
#include "helpserv.h"
#include "hybdefs.h"
#include "log.h"
#include "match.h"
#include "memoserv.h"
#include "misc.h"
#include "mystring.h"
#include "nickserv.h"
#include "hostserv.h"
#include "operserv.h"
#include "settings.h"
#include "sock.h"
#include "timestr.h"
#include "sprintf_irc.h"
#include "server.h"

#ifdef NICKSERVICES

/* 
 * hash containing registered nick info
 */
struct NickInfo *nicklist[NICKLIST_MAX];

static int myi;

static int ChangePass(struct NickInfo *, char *, char *);
static int AddHostToNick(char *, struct NickInfo *);

#ifdef HOSTSERVICES
static int AddVHostToNick(char *, time_t ts, struct NickInfo *);
#endif /* HOSTSERVICES */

static struct NickInfo *MakeNick();
static void AddNick(struct NickInfo *);
static int OnAccessList(char *, char *, struct NickInfo *);

#ifdef LINKED_NICKNAMES
static int InsertLink(struct NickInfo *hub, struct NickInfo *leaf);
static int DeleteLink(struct NickInfo *nptr, int copyhosts);
#endif /* LINKED_NICKNAMES */

static void n_help(struct Luser *, int, char **);
static void n_register(struct Luser *, int, char **);
static void n_drop(struct Luser *, int, char **);
static void n_identify(struct Luser *, int, char **);
static void n_recover(struct Luser *, int, char **);
static void n_release(struct Luser *, int, char **);
static void n_ghost(struct Luser *, int, char **);

static void n_access(struct Luser *, int, char **);
static void n_access_add(struct Luser *, struct NickInfo *, int, char **);
static void n_access_del(struct Luser *, struct NickInfo *, int, char **);
static void n_access_list(struct Luser *, struct NickInfo *, int, char **);

static void n_set(struct Luser *, int, char **);
static void n_set_kill(struct Luser *, int, char **);
static void n_set_automask(struct Luser *, int, char **);
static void n_set_private(struct Luser *, int, char **);
static void n_set_oper(struct Luser *, int, char **);
static void n_set_secure(struct Luser *, int, char **);
static void n_set_unsecure(struct Luser *, int, char **);
static void n_set_memos(struct Luser *, int, char **);
static void n_set_notify(struct Luser *, int, char **);
static void n_set_helplang(struct Luser *, int, char **);
static void n_set_signon(struct Luser *, int, char **);
static void n_set_hide(struct Luser *, int, char **);
static void n_set_password(struct Luser *, int, char **);
static void n_set_email(struct Luser *, int, char **);
static void n_set_url(struct Luser *, int, char **);
static void n_set_uin(struct Luser *, int, char **);
static void n_set_gsm(struct Luser *, int, char **);
static void n_set_phone(struct Luser *, int, char **);
#ifdef LINKED_NICKNAMES
static void n_set_master(struct Luser *, int, char **);
#endif

static void n_list(struct Luser *, int, char **);
static void n_info(struct Luser *, int, char **);

#ifdef LINKED_NICKNAMES
static void n_link(struct Luser *, int, char **);
static void n_unlink(struct Luser *, int, char **);
#endif /* LINKED_NICKNAMES */

static void n_fixts(struct Luser *, int, char **);

#ifdef EMPOWERADMINS

static void n_forbid(struct Luser *, int, char **);
static void n_unforbid(struct Luser *, int, char **);
static void n_setpass(struct Luser *, int, char **);
static void n_noexpire(struct Luser *, int, char **);

#ifdef LINKED_NICKNAMES
static void n_showlink(struct Luser *, int, char **);
static void n_droplink(struct Luser *, int, char **);
#endif /* LINKED_NICKNAMES */

static void n_collide(struct Luser *, int, char **);
static void n_flag(struct Luser *, int, char **);
static void n_clearnoexp(struct Luser *, int, char **);

#endif /* EMPOWERADMINS */

/* main NickServ commands */
static struct Command nickcmds[] = {
  { "HELP", n_help, LVL_NONE },
  { "REGISTER", n_register, LVL_NONE },
  { "DROP", n_drop, LVL_NONE },
  { "IDENTIFY", n_identify, LVL_NONE },
  { "RECOVER", n_recover, LVL_NONE },
  { "RELEASE", n_release, LVL_NONE },
  { "GHOST", n_ghost, LVL_NONE },
  { "ACCESS", n_access, LVL_IDENT },
  { "SET", n_set, LVL_IDENT },
  { "LIST", n_list, LVL_NONE },
  { "INFO", n_info, LVL_NONE },

#ifdef LINKED_NICKNAMES
  { "LINK", n_link, LVL_IDENT },
  { "UNLINK", n_unlink, LVL_IDENT },
#endif /* LINKED_NICKNAMES */

#ifdef EMPOWERADMINS

  { "FORBID", n_forbid, LVL_ADMIN },
  { "UNFORBID", n_unforbid, LVL_ADMIN },
  { "SETPASS", n_setpass, LVL_ADMIN },
  { "CHPASS", n_setpass, LVL_ADMIN },
  { "NOEXPIRE", n_noexpire, LVL_ADMIN },
#ifdef LINKED_NICKNAMES
  { "SHOWLINK", n_showlink, LVL_ADMIN },
  { "DROPLINK", n_droplink, LVL_ADMIN },
#endif /* LINKED_NICKNAMES */
  { "COLLIDE", n_collide, LVL_ADMIN },
  { "FLAG", n_flag, LVL_ADMIN },
  { "CLEARNOEXP", n_clearnoexp, LVL_ADMIN },
  { "FIXTS", n_fixts, LVL_ADMIN },

#endif /* EMPOWERADMINS */

  { 0, 0, 0 }
};

/* sub commands of NickServ ACCESS */
static struct Command accesscmds[] = {
  { "ADD", n_access_add, LVL_NONE },
  { "DEL", n_access_del, LVL_NONE },
  { "LIST", n_access_list, LVL_NONE },

  { 0, 0, 0 }
};

/* sub commands of NickServ SET */
static struct Command setcmds[] = {
  { "KILL", n_set_kill, LVL_NONE },
  { "AUTOMASK", n_set_automask, LVL_NONE },
  { "PRIVATE", n_set_private, LVL_NONE },
  { "OPER", n_set_oper, LVL_NONE },
  { "SECURE", n_set_secure, LVL_NONE },
  { "UNSECURE", n_set_unsecure, LVL_NONE },
  { "MEMOS", n_set_memos, LVL_NONE },
  { "NOTIFY", n_set_notify, LVL_NONE },
  { "HELPLANG", n_set_helplang, LVL_NONE },
  { "SIGNON", n_set_signon, LVL_NONE },
  { "HIDE", n_set_hide, LVL_NONE },
  { "PASSWORD", n_set_password, LVL_NONE },
  { "EMAIL", n_set_email, LVL_NONE },
  { "URL", n_set_url, LVL_NONE },
  { "UIN", n_set_uin, LVL_NONE },
  { "GSM", n_set_gsm, LVL_NONE },
  { "PHONE", n_set_phone, LVL_NONE },

#ifdef LINKED_NICKNAMES
  { "MASTER", n_set_master, LVL_NONE },
#endif

  { 0, 0, 0 }
};

/*
ns_process()
  Process command coming from 'nick' directed towards n_NickServ
*/

void
ns_process(char *nick, char *command)

{
  int acnt;
  char **arv;
  struct Command *cptr;
  struct Luser *lptr;
  struct NickInfo *nptr;

  if (!command)
    return;

  lptr = FindClient(nick);
  if (!lptr)
    return;

  if (Network->flags & NET_OFF)
  {
    notice(n_NickServ, lptr->nick,
      "Services are currently \002disabled\002");
    return;
  }

  acnt = SplitBuf(command, &arv);
  if (acnt == 0)
  {
    MyFree(arv);
    return;
  }

  cptr = GetCommand(nickcmds, arv[0]);

  if (!cptr || (cptr == (struct Command *) -1))
  {
    notice(n_NickServ, lptr->nick,
      "%s command [%s]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      arv[0]);
    MyFree(arv);
    return;
  }

  /*
   * Check if the command is for admins only - if so,
   * check if they match an admin O: line, and are
   * currently registered with OperServ.
   */
  if ((cptr->level == LVL_ADMIN) && !(IsValidAdmin(lptr)))
  {
    notice(n_NickServ, lptr->nick, "Unknown command [%s]",
      arv[0]);
    MyFree(arv);
    return;
  }

  /*
   * Use FindNick() here instead of GetLink() because
   * nicknames are identified individually, and using GetLink()
   * would tell us if the master of the link is identified,
   * not necessarily this specific nickname
   */
  if ((nptr = FindNick(lptr->nick)))
    if (nptr->flags & NS_FORBID)
    {
      notice(n_NickServ, lptr->nick,
        "Cannot execute commands for forbidden nicknames");
      MyFree(arv);
      return;
    }

  if (cptr->level != LVL_NONE)
  {
    if (!nptr)
    {
      notice(n_NickServ, lptr->nick,
        "You must register your nick with \002%s\002 first",
        n_NickServ);
      notice(n_NickServ, lptr->nick,
        "Type: \002/msg %s REGISTER <password>\002 to register your nickname", 
        n_NickServ);
      MyFree(arv);
      return;
    }

    if (!(nptr->flags & NS_IDENTIFIED))
    {
      notice(n_NickServ, lptr->nick,
        "Password identification is required for [\002%s\002]",
        cptr->cmd);
      notice(n_NickServ, lptr->nick, 
        "Type \002/msg %s IDENTIFY <password>\002 and retry",
        n_NickServ);
      MyFree(arv);
      return;
    }
  }

  /* call cptr->func to execute command */
  (*cptr->func)(lptr, acnt, arv);

  MyFree(arv);

  return;
} /* ns_process() */

/*
ns_loaddata()
  Load NickServ database - return 1 if successful, -1 if not, and -2
if the errors are unrecoverable
*/

int
ns_loaddata()

{
  FILE *fp;
  char line[MAXLINE], **av;
  char *keyword;
  int ac, ret = 1, cnt;
  int islink;
  struct NickInfo *nptr = NULL;

  if (!(fp = fopen(NickServDB, "r")))
  {
    /* NickServ data file doesn't exist */
    return -1;
  }

  cnt = 0;
  islink = 0;

  /* load data into list */
  while (fgets(line, MAXLINE - 1, fp))
  {
    cnt++;
    ac = SplitBuf(line, &av);
    if (!ac)
    {
      /* probably a blank line */
      MyFree(av);
      continue;
    }

    if (av[0][0] == ';')
    {
      MyFree(av);
      continue;
    }

    if (!ircncmp("->", av[0], 2))
    {
      /* 
       * check if there are enough args
       */
      if (ac < 2)
      {
        fatal(1, "%s:%d Invalid database format (FATAL)",
          NickServDB,
          cnt);
        ret = -2;
        continue;
      }

      /* check if there is no nickname associated with the data */
      if (!nptr)
      {
        fatal(1, "%s:%d No nickname associated with data",
          NickServDB,
          cnt);
        if (ret > 0)
          ret = -1;
        continue;
      }

      keyword = av[0] + 2;
      if (!ircncmp(keyword, "PASS", 4))
      {
        if (!nptr->password)
		{
          nptr->password = MyStrdup(av[1]);
	  	  nptr->securepass = MyStrdup(ac>2? av[2] : av[1]); 
        }
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple PASS lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!ircncmp(keyword, "HOST", 4) && !islink)
      {
        AddHostToNick(av[1], nptr);
      }
      
    #ifdef HOSTSERVICES
      
      else if (!ircncmp(keyword, "SPOOF", 5) && !islink)
      {
        AddVHostToNick(av[1], atol(av[2]), nptr);
      }
      
      else if (!ircncmp(keyword, "CHANGES", 7) && !islink)
      {
		int i;
		
		for (i=0; i < ac-1; i++)
			nptr->vhchanges[i] = atol(av[i+1]);
      }
      
    #endif /* HOSTSERVICES */
      
      else if (!ircncmp(keyword, "EMAIL", 5))
      {
        if (!nptr->email)
          nptr->email = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple EMAIL lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!ircncmp(keyword, "URL", 3))
      {
        if (!nptr->url)
          nptr->url = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple URL lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!ircncmp(keyword, "GSM", 3))
      {
        if (!nptr->gsm)
          nptr->gsm = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple GSM lines (using first)",
            NickServDB, cnt, nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!ircncmp(keyword, "PHONE", 5))
      {
        if (!nptr->phone)
          nptr->phone = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple PHONE lines (using first)",
            NickServDB, cnt, nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!ircncmp(keyword, "UIN", 3))
      {
        if (!nptr->UIN)
          nptr->UIN = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple UIN lines (using first)",
            NickServDB, cnt, nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (LastSeenInfo && !ircncmp(keyword, "LASTUH", 6))
      {
        if (!nptr->lastu && !nptr->lasth)
        {
          nptr->lastu = MyStrdup(av[1]);
          nptr->lasth = MyStrdup(av[2]);
        }
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple LASTUH lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (LastSeenInfo && !ircncmp(keyword, "LASTQMSG", 6))
      {
        if (!nptr->lastqmsg)
          nptr->lastqmsg = MyStrdup(av[1] + 1);
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple LASTQMSG lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }

    #ifdef LINKED_NICKNAMES

      else if (!ircncmp(keyword, "LINK", 4))
      {
        if (!nptr->master)
        {
          struct NickInfo *master;

          if (!(master = GetLink(av[1])))
          {
            fatal(1, "%s:%d NickServ entry for [%s] has an unregistered master nickname: %s",
              NickServDB,
              cnt,
              nptr->nick,
              av[1]);
            if (ret > 0)
              ret = -1;
          }
          else
          {
            int    goodlink;

            /*
             * The master nickname is ok - insert nptr into
             * nptr->master's link list
             */
            islink = 1;

            goodlink = InsertLink(master, nptr);
            if (goodlink <= 0)
            {
              fatal(1, "%s:%d InsertLink() failed for nickname [%s]: %s (FATAL)",
                NickServDB,
                cnt,
                nptr->nick,
                (goodlink == -1) ? "Circular Link" : "Exceeded MaxLinks links");
              ret = -2;
            }
            else
            {
              struct NickHost *hptr;

              /* Delete nptr's access list */
              while (nptr->hosts)
              {
                hptr = nptr->hosts->next;
                MyFree(nptr->hosts->hostmask);
                MyFree(nptr->hosts);
                nptr->hosts = hptr;
              }
            }
          }
        }
        else
        {
          fatal(1, "%s:%d NickServ entry for [%s] has multiple LINK lines (using first)",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }
      }

    #endif /* LINKED_NICKNAMES */

    } /* if (!ircncmp("->", keyword, 2)) */
    else
    {
      if (nptr)
      {
        if (!nptr->password && !(nptr->flags & NS_FORBID))
        {
          /* the previous nick didn't have a PASS line */
          fatal(1,
            "%s:%d No password entry for registered nick [%s] (FATAL)",
            NickServDB,
            cnt,
            nptr->nick);
          ret = -2;
        }

        if (!nptr->hosts && !(nptr->flags & NS_FORBID) && !islink)
        {
          /* the previous nick didn't have a HOST line */
          fatal(1, "%s:%d No hostname entry for registered nick [%s]",
            NickServDB,
            cnt,
            nptr->nick);
          if (ret > 0)
            ret = -1;
        }

        /*
         * we've come to a new nick entry, so add the last nick
         * to the list before proceeding
         */
        AddNick(nptr);

        if (islink)
          islink = 0;
      }

      /* 
       * make sure there are enough args on the line
       * <nick> <flags> <time created> <last seen>
       */
      if (ac < 4)
      {
        fatal(1, "%s:%d Invalid database format (FATAL)",
          NickServDB,
          cnt);
        ret = -2;
        nptr = NULL;
        continue;
      }
#if 0
      /* Check if there already exists that nickname in list. This will
       * give some overhead, but this will make sure no nicknames are
       * twice or more times in db. */
      if (FindNick(av[0]))
      {
        fatal(1, "%s:%d NickServ entry [%s] is already in nick list",
            NickServDB, cnt, av[0]);
        nptr = NULL;
        ret = -1;
      }
      else
      {
#endif
        nptr = MakeNick();
        nptr->nick = MyStrdup(av[0]);
        nptr->flags = atol(av[1]);
        nptr->created = atol(av[2]);
        nptr->lastseen = atol(av[3]);
        nptr->flags &= ~NS_IDENTIFIED;
#if 0
      }
#endif
    }
    MyFree(av);
  } /* while */

  /*
   * This is needed because the above loop will only add the nicks
   * up to, but not including, the last nick in the database, so
   * we need to add it now - Also, if there is only 1 nick entry
   * it will get added now
   */
  if (nptr)
  {
    if ((nptr->password) || (nptr->flags & NS_FORBID))
    {
      if (!nptr->hosts && !(nptr->flags & NS_FORBID) && !islink)
      {
        fatal(1, "%s:%d No hostname entry for registered nick [%s]",
          NickServDB,
          cnt,
          nptr->nick);
        if (ret > 0)
          ret = -1;
      }

      /* 
       * nptr has a password, so it can be added to the table
       */
      AddNick(nptr);
    }
    else
    {
      if (!nptr->password && !(nptr->flags & NS_FORBID))
      {
        fatal(1, "%s:%d No password entry for registered nick [%s] (FATAL)",
          NickServDB,
          cnt,
          nptr->nick);
        ret = -2;
      }
    }
  } /* if (nptr) */

  fclose(fp);

  return ret;
} /* ns_loaddata */

/*
AddHostToNick()
  Add 'host' to list of hosts for 'nickptr'
*/

static int
AddHostToNick(char *host, struct NickInfo *nickptr)

{
  int i;
  struct NickHost *hptr;

  if (!nickptr || !host)
    return 1;

  for (hptr = nickptr->hosts, i = 0; hptr; hptr = hptr->next, ++i);

  if (MaxNickHosts && i >= MaxNickHosts)
    return 0;

  hptr = (struct NickHost *) MyMalloc(sizeof(struct NickHost));
  hptr->hostmask = MyStrdup(host);
  hptr->next = nickptr->hosts;
  nickptr->hosts = hptr;

  return 1;
} /* AddHostToNick() */

#ifdef HOSTSERVICES

static int
AddVHostToNick(char *host, time_t ts, struct NickInfo *nickptr)

{
  int i;
  struct NickVHost *vptr;

  if (!nickptr || !host)
    return 1;

  for (vptr = nickptr->vhosts, i = 0; vptr; vptr = vptr->next, ++i);

  if (VhostPerNick && i >= VhostPerNick)
    return 0;

  vptr = (struct NickVHost *) MyMalloc(sizeof(struct NickVHost));
  vptr->hostmask = MyStrdup(host);
  vptr->next = nickptr->vhosts;
  nickptr->vhosts = vptr;
  
  vptr->ts = ts;

  return 1;
} /* AddVHostToNick() */

#endif /* HOSTSERVICES */

#ifdef CHANNELSERVICES

/*
AddFounderChannelToNick()
 Each nick structure contains a list of channels for which they
are registered as a founder (ie: they created the channel). Add
such a channel to 'nptr's list
*/

void
AddFounderChannelToNick(struct NickInfo **nptr, struct ChanInfo *cptr)

{
  struct aChannelPtr *tmp;

  tmp = (struct aChannelPtr *) MyMalloc(sizeof(struct aChannelPtr));
  memset(tmp, 0, sizeof(struct aChannelPtr));

  tmp->cptr = cptr;
  tmp->next = (*nptr)->FounderChannels;
  (*nptr)->FounderChannels = tmp;

  ++(*nptr)->fccnt;
} /* AddFounderChannelToNick() */

/*
 * RemoveFounderChannelFromNick()
 *
 * Remove 'cptr' from 'nptr's founder channel list
*/

void RemoveFounderChannelFromNick(struct NickInfo **nptr, struct ChanInfo
    *cptr)
{
  struct aChannelPtr *tmp, *prev;

  prev = NULL;

  /* Iterate list of nick's founder channels */
  for (tmp = (*nptr)->FounderChannels; tmp; )
  {
    /* We have a match! */
    if (tmp->cptr == cptr)
    {
      /* Last time we didn't have a match, so we have a ptr before this
       * one in case of list relinking */
      if (prev)
      {
        prev->next = tmp->next;
        MyFree(tmp);
        tmp = prev;
      }
      else
      {
        (*nptr)->FounderChannels = tmp->next;
        MyFree(tmp);
        tmp = NULL;
      }

      /* KrisDuv's fix for decreasing number of registered channels */
      --(*nptr)->fccnt;

      /* We can break since there should always be only 1 match */
      break;
    }

    prev = tmp;

    if (tmp)
      tmp = tmp->next;
    else
      tmp = (*nptr)->FounderChannels;
  }
} /* RemoveFounderChannelFromNick() */

/*
AddAccessChannel()
 Add an access channel entry for nptr. Return new entry
*/

struct AccessChannel *
AddAccessChannel(struct NickInfo *nptr, struct ChanInfo *chanptr,
                 struct ChanAccess *accessptr)

{
  struct AccessChannel *acptr;

  acptr = (struct AccessChannel *) MyMalloc(sizeof(struct AccessChannel));
  acptr->cptr = chanptr;
  acptr->accessptr = accessptr;

  acptr->prev = NULL;
  acptr->next = nptr->AccessChannels;
  if (acptr->next)
    acptr->next->prev = acptr;
  nptr->AccessChannels = acptr;

  return (acptr);
} /* AddAccessChannel() */

/*
DeleteAccessChannel()
 Delete access channel entry for nptr
*/

void
DeleteAccessChannel(struct NickInfo *nptr, struct AccessChannel *acptr)

{
  assert(nptr && acptr);

  if (acptr->next)
    acptr->next->prev = acptr->prev;
  if (acptr->prev)
    acptr->prev->next = acptr->next;
  else
    nptr->AccessChannels = acptr->next;

  MyFree(acptr);
} /* DeleteAccessChannel() */

#endif /* CHANNELSERVICES */

/*
MakeNick()
 Allocate memory for a NickInfo structure and initialize
its fields
*/

static struct NickInfo *
MakeNick()

{
  struct NickInfo *nptr;

  nptr = (struct NickInfo *) MyMalloc(sizeof(struct NickInfo));

  memset(nptr, 0, sizeof(struct NickInfo));

  return (nptr);
} /* MakeNick() */

/*
AddNick()
  Add 'nickptr' to nickname table
*/

static void
AddNick(struct NickInfo *nickptr)

{
  unsigned int hashv;

  /* 
   * all the fields of nickptr were filled in already, so just 
   * insert it in the list
   */
  hashv = NSHashNick(nickptr->nick);

  nickptr->prev = NULL;
  nickptr->next = nicklist[hashv];
  if (nickptr->next)
    nickptr->next->prev = nickptr;
  nicklist[hashv] = nickptr;

  ++Network->TotalNicks;
} /* AddNick() */

/*
DeleteNick()
 Delete structure 'nickptr' - assume calling function handles
list manipulation - setting prev/next variables etc.
*/

void
DeleteNick(struct NickInfo *nickptr)

{
  struct NickHost *htmp;
  unsigned int hashv;
#ifdef CHANNELSERVICES
  struct ChanInfo *cptr;
  struct aChannelPtr *ftmp;
  struct AccessChannel *atmp;
#endif
#ifdef MEMOSERVICES
  struct MemoInfo *mi;
#endif

  if (!nickptr)
    return;

#ifdef MEMOSERVICES
  /* check if the nick had any memos - if so delete them */
  if ((mi = FindMemoList(nickptr->nick)))
    DeleteMemoList(mi);
#endif

#ifdef LINKED_NICKNAMES
  DeleteLink(nickptr, 0);
#endif /* LINKED_NICKNAMES */

  hashv = NSHashNick(nickptr->nick);

  MyFree(nickptr->nick);
  MyFree(nickptr->password);
  MyFree(nickptr->securepass);

  while (nickptr->hosts)
  {
    htmp = nickptr->hosts->next;
    MyFree(nickptr->hosts->hostmask);
    MyFree(nickptr->hosts);
    nickptr->hosts = htmp;
  }

#ifdef HOSTSERVICES

  while (nickptr->vhosts)
  {
    struct NickVHost *vtmp;
  
    vtmp = nickptr->vhosts->next;
    MyFree(nickptr->vhosts->hostmask);
    MyFree(nickptr->vhosts);
    nickptr->vhosts = vtmp;
  }
  
  if (nickptr->authcookie)
  	MyFree(nickptr->authcookie);

#endif /* HOSTSERVICES */

#ifdef CHANNELSERVICES

  while (nickptr->FounderChannels)
  {
    cptr = nickptr->FounderChannels->cptr;

    ftmp = nickptr->FounderChannels->next;
    MyFree(nickptr->FounderChannels);
    nickptr->FounderChannels = ftmp;

    /*
     * All channels that this nick registered should be dropped, unless
     * there is a successor.
     *
     * Before calling DeleteChan(), it would be best if nickptr has
     * already been removed from nicklist[]. Otherwise, DeleteChan() might
     * try to call RemoveFounderChannelFromNick(). This would be very bad,
     * since this loop is modifying nickptr's FounderChannels list. Right
     * now, all calls to DeleteNick() remove nickptr from nicklist[]
     * beforehand, but, being as paranoid as I am, we'll set cptr->founder
     * to null here, so there is *NO* chance of it ever being used to
     * delete nickptr's FounderChannels list.
     */
    if (cptr->founder)
    {
      MyFree(cptr->founder);
      cptr->founder = NULL;
    }

    /*
     * If the channel has a successor, promote them to founder,
     * otherwise delete the channel
     */
    if (cptr->successor)
      PromoteSuccessor(cptr);
    else
    {
      /* Fix by KrisDuv - make ChanServ part if on channel */
      struct Channel *chptr;
      chptr = FindChannel(cptr->name);
      if (IsChannelMember(chptr, Me.csptr))
        cs_part(chptr);
       
      /* And delete channel finally */   
      DeleteChan(cptr);
    }
  }

  /*
   * remove nickptr from ACCESS/AKICK lists modifier/creator field
   */
  for (atmp = nickptr->AccessChannels; atmp; atmp = atmp->next)
    {
      struct ChanAccess *acl;
      struct AutoKick *akick;

      for (acl = atmp->cptr->access; acl; acl = acl->next)
	if (acl->mptr == nickptr)
	  acl->mptr = 0;

      for (akick = atmp->cptr->akick; akick; akick = akick->next)
	if (akick->who == nickptr)
	  akick->who = 0;
    } 

  /*
   * Before deleting nickptr->AccessChannels, we first want
   * to parse the list and delete the ChanAccess structures
   * which point to nickptr - otherwise another person could
   * register the same nickname and gain instant access to
   * the channels the old nick had access on.
   */
  for (atmp = nickptr->AccessChannels; atmp; atmp = atmp->next)
    DeleteAccess(atmp->cptr, atmp->accessptr);

  while (nickptr->AccessChannels)
  {
    atmp = nickptr->AccessChannels->next;
    MyFree(nickptr->AccessChannels);
    nickptr->AccessChannels = atmp;
  }

#endif /* CHANNELSERVICES */

  if (nickptr->email)
    MyFree(nickptr->email);
  if (nickptr->url)
    MyFree(nickptr->url);
  if (nickptr->lastu)
    MyFree(nickptr->lastu);
  if (nickptr->lasth)
    MyFree(nickptr->lasth);
  if (nickptr->lastqmsg)
    MyFree(nickptr->lastqmsg);

  if (nickptr->next)
    nickptr->next->prev = nickptr->prev;
  if (nickptr->prev)
    nickptr->prev->next = nickptr->next;
  else
    nicklist[hashv] = nickptr->next;

  MyFree(nickptr);

  --Network->TotalNicks;
} /* DeleteNick() */

/*
HasFlag()
 Determine whether 'nick' has the flag 'flag'
*/

int
HasFlag(char *nick, int flag)

{
  struct NickInfo *master;

  if ((master = GetLink(nick)))
    return (master->flags & flag);

  return (0);
} /* HasFlag() */

/*
ChangePass
  Change password for nptr->nick to 'newpass'
*/

static int
ChangePass(struct NickInfo *nptr, char *newpass, char *newsecure)

{
#ifdef CRYPT_PASSWORDS
  char *encr;
#endif

  if (!nptr || !newpass)
    return 0;

  MyFree(nptr->password);
  MyFree(nptr->securepass);

#ifdef CRYPT_PASSWORDS

  /* encrypt it */
  encr = hybcrypt(newpass, NULL);
  assert(encr != 0);

  nptr->password = MyStrdup(encr);

  encr = hybcrypt(newsecure, NULL);
  assert(encr != 0);

  nptr->securepass = MyStrdup(encr);

#else

  /* just use plaintext */
  nptr->password = MyStrdup(newpass);
  nptr->securepass = MyStrdup(newsecure);

#endif

  return 1;
} /* ChangePass() */

/*
CheckNick()
  If 'nickname' 's host is recognized, return 1; if 'nickname' is
not registered, or does not match a known host, return 0 and
prepare to collide 'nickname'
*/

int
CheckNick(char *nickname)

{
  struct Luser *lptr;
  struct NickInfo *nptr, *realptr;
  int knownhost;

  realptr = FindNick(nickname);
  nptr = GetMaster(realptr);
  if (!realptr || !nptr)
    return 0; /* nickname is not registered */

  /*
   * If RECORD_SPLIT_TS is defined, the user could possibly be
   * identified already, if we are rejoining from a netsplit;
   * if that is the case, don't bother checking hostnames etc
   */
  if (realptr->flags & NS_IDENTIFIED)
    return 0;

  if (!(lptr = FindClient(nickname)))
    return 0; /* nickname doesn't exist */

  /*
   * Check if the nick is forbidden
   */
  if (realptr->flags & NS_FORBID)
  {
    notice(n_NickServ, lptr->nick, 
      "This nickname may not be used.  Please choose another.");
#ifdef BNET
    notice(n_NickServ, lptr->nick,
      "If you do not change within one minute, your nick will be forcefully changed");
#else
    notice(n_NickServ, lptr->nick,
      "If you do not change within one minute, you will be disconnected");
#endif
    realptr->flags |= NS_COLLIDE;
    realptr->collide_ts = current_ts + 60;
    return 0;
  }

  /*
   * We don't need to call OnAccessList() if the nickname
   * is secure - waste of cpu time
   */
  if (nptr->flags & NS_SECURE)
    knownhost = 0;
  else
    knownhost = OnAccessList(lptr->username, lptr->hostname, nptr);

  /*
   * If knownhost is 0, then we know either the nickname is
   * secure, or their hostname is not on their access list -
   * in either case, give them a warning
   */
  if (!knownhost)
  {
    if (NicknameWarn || (nptr->flags & NS_PROTECTED))
    {
      /*
       * If NicknameWarn is enabled, or if their nickname
       * is kill protected, give them a warning telling
       * them to identify
       */
      notice(n_NickServ, lptr->nick, ERR_NOT_YOUR_NICK);
      notice(n_NickServ, lptr->nick, ERR_NEED_IDENTIFY, n_NickServ);
    }

   if (AllowKillProtection)
   {
    if (nptr->flags & NS_KILLIMMED)
    {
      if (nptr->flags & NS_SECURE)
      {
        /*
         * If the nickname is secure, we cannot kill them
         * immediately, because they will be killed every time
         * they logon, whether they come from a known host or not.
         * Give them one minute to identify.
         */
        notice(n_NickServ, lptr->nick, ERR_MUST_CHANGE);
        realptr->flags |= NS_COLLIDE;
        realptr->collide_ts = current_ts + 60;
      }
      else
      {
        /*
          * The nickname has kill immediately set, so collide
          * the nickname right now
          */
        putlog(LOG1, 
          "%s: immediately killing %s!%s@%s (Nickname Enforcement)",
          n_NickServ,
          lptr->nick,
          lptr->username,
          lptr->hostname);

        SendUmode(OPERUMODE_S,
          "%s: immediately killing %s!%s@%s (Nickname Enforcement)",
          n_NickServ,
          lptr->nick,
          lptr->username,
          lptr->hostname);

        collide(lptr->nick);

        nptr->collide_ts = 0;
      }
    }
    else if (nptr->flags & NS_PROTECTED)
    {
      /*
       * If the nick is kill protected, give a 1 minute timeout
       * for them to change
       */
      notice(n_NickServ, lptr->nick, ERR_MUST_CHANGE);
      realptr->flags |= NS_COLLIDE;
      realptr->collide_ts = current_ts + 60;
    }
   }
   return 0;
  } /* if (!knownhost) */

  if ((knownhost) && (nptr->flags & NS_UNSECURE))
  {
    /*
     * They're from a known host, and have UNSECURE set -
     * mark them as identified
     */
    realptr->flags |= NS_IDENTIFIED;
#ifdef BNET
  toserv(":%s SVSHINT 1 :%s\n", Me.name, lptr->nick);
#endif
  }
  return 1;
} /* CheckNick() */

/*
CheckOper()
  Check if operator 'nickname' has set the OPER flag for their
nickname - we know 'nickname' is already an operator
*/

void
CheckOper(char *nickname)

{
  struct NickInfo *nptr, *realptr;

  if (!nickname)
    return;

  realptr = FindNick(nickname);
  nptr = GetMaster(realptr);

  if (realptr && nptr)
    if ((realptr->flags & NS_IDENTIFIED) && !(nptr->flags & NS_OPERATOR))
    {
      notice(n_NickServ, nickname,
        "You have not set the IRC Operator flag for your nickname");
      notice(n_NickServ, nickname,
        "Please type \002/msg %s SET OPER ON\002 so your nickname does not expire",
        n_NickServ);
    }
} /* CheckOper() */

/*
ExpireNicknames()
 Go through nickname list and delete any nicknames that
have expired.
*/

void
ExpireNicknames(time_t unixtime)

{
  int ii;
  struct NickInfo *nptr, *next;

  for (ii = 0; ii < NICKLIST_MAX; ++ii)
  {
    for (nptr = nicklist[ii]; nptr; nptr = next)
    {
      next = nptr->next;

    #ifdef RECORD_SPLIT_TS

      if (nptr->split_ts)
      {
        /*
         * Reset nptr->split_ts every half hour, if we have not
         * seen them
         */
        if ((unixtime - nptr->whensplit) >= 1800)
          nptr->split_ts = nptr->whensplit = 0;
      }

    #endif /* RECORD_SPLIT_TS */

      if (NickNameExpire)
      {
        if ((!(nptr->flags & (NS_FORBID | NS_OPERATOR | NS_IDENTIFIED)))
            && ((unixtime - nptr->lastseen) >= NickNameExpire))
        {
          putlog(LOG2,
            "%s: Expired nickname: [%s]",
            n_NickServ,
            nptr->nick);

          DeleteNick(nptr);
        }
      } /* if (NickNameExpire) */
    } /* for (nptr = nicklist[ii]; nptr; nptr = next) */
  } /* for (ii = 0; ii < NICKLIST_MAX; ++ii) */
} /* ExpireNicknames() */

/*
FindNick()
  Return a pointer to registered nick 'nickname'
*/

struct NickInfo *
FindNick(char *nickname)

{
  struct NickInfo *nptr;
  unsigned int hashv;

  if (!nickname)
    return (NULL);

  hashv = NSHashNick(nickname);
  for (nptr = nicklist[hashv]; nptr; nptr = nptr->next)
    if (!irccmp(nptr->nick, nickname))
      return (nptr);

  return (NULL);
} /* FindNick() */

/*
GetLink()
 Similar to FindNick(), except return a pointer to nickname's
master if nickname is in a nick link
*/

struct NickInfo *
GetLink(char *nickname)

{
  struct NickInfo *nptr;

  if (!(nptr = FindNick(nickname)))
    return (NULL);

#ifdef LINKED_NICKNAMES
  if (nptr->master)
    return (nptr->master);
#endif /* LINKED_NICKNAMES */

  return (nptr);
} /* GetLink() */

/*
OnAccessList()
  Return 1 if hostmask is on the list of recognized hosts
for registered nick 'nptr'; 0 if not
*/

static int
OnAccessList(char *username, char *hostname, struct NickInfo *nptr)

{
  struct NickHost *hptr;
  char hostmask[UHOSTLEN + 2];

  if (!username || !hostname || !nptr)
    return 0;

  ircsprintf(hostmask, "%s@%s", username, hostname);

  for (hptr = nptr->hosts; hptr; hptr = hptr->next)
    if (match(hptr->hostmask, hostmask))
      return 1;

  return 0;
} /* OnAccessList() */

/*
collide()
  Send a nick collide for 'nick'
*/

void
collide(char *nick)

{
  struct Luser *lptr;
  struct NickInfo *nptr;
  char sendstr[MAXLINE];
#ifndef BNET
  char **av;
#endif 

  if( !SafeConnect )
     return;

  if (!(lptr = FindClient(nick)))
    return;

  ircsprintf(sendstr,
  #ifdef BNET
    "NICK %s 1 %ld +i %s %s %s %s :%s\n",
  #else
    "NICK %s 1 %ld +i %s %s %s :%s\n",
  #endif
    lptr->nick,
/*  (long) (lptr->nick_ts - 1),*/
    (long) time(NULL),
    "enforced",
    Me.name,
    Me.name,
  #ifdef BNET
    Me.name,
  #endif
    "Nickname Enforcement");

  /*
   * Sending a server kill will be quieter than an oper
   * kill since most clients are -k
   */

#ifdef BNET
  if (!IsIgnoredServer(lptr->server->name))
    toserv(":%s SVSNICK %s %s :%s\n", 
	 Me.name, 
	 lptr->server->name,
	 lptr->nick,
	 getGuest());
  else
    toserv("KILL %s :%s!%s (Nickname Enforcement on Ghost)\n",
	   lptr->nick, 
	   Me.name,
	   n_NickServ);

  lptr->nick_ts = time(NULL);
#else
  toserv("KILL %s :%s!%s (Nickname Enforcement)\n%s",
    lptr->nick,
    Me.name,
    n_NickServ,
    sendstr);

  /* erase the old user */
  DeleteClient(lptr);

  /* add the new one */
  SplitBuf(sendstr, &av);
  AddClient(av);

  MyFree(av);
#endif

  if ((nptr = FindNick(nick)))
  {
    /*
     * remove the collide timer, but put a release timer so the
     * pseudo nick gets removed after NSReleaseTimeout
     */
    nptr->flags &= ~NS_COLLIDE;
    nptr->flags |= NS_RELEASE;
#ifdef BNET
    nptr->flags |= NS_SVSNICK;
#endif
  }
} /* collide() */

/*
release()
  Release the enforced nickname 'nickname'
*/

void
release(char *nickname)

{
#ifndef BNET
  struct Luser *lptr;
#endif
  struct NickInfo *nptr;

#ifdef BNET
  if (!(nptr = FindNick(nickname)))
    return;

  toserv(":%s QUIT :Released\n", nptr->nick);
#else
  if (!(lptr = FindClient(nickname)) || !(nptr = FindNick(nickname)))
    return;

  if (!lptr->server)
    return;

  if (lptr->server != Me.sptr)
    return; /* lptr->nick isn't a pseudo-nick */

  toserv(":%s QUIT :Released\n",
    lptr->nick);
  DeleteClient(lptr);
#endif

  nptr->flags &= ~NS_RELEASE;
#ifdef BNET
  nptr->flags &= ~NS_SVSNICK;
#endif
} /* release() */

/*
CollisionCheck()
 Called from DoTimer() to nick collide anyone who hasn't
identified by the TS given by nptr->collide_ts, or release
nicknames that have passed the release timeout.
*/

void
CollisionCheck(time_t unixtime)

{
  int ii;
  struct NickInfo *nptr;

  for (ii = 0; ii < NICKLIST_MAX; ii++)
  {
    for (nptr = nicklist[ii]; nptr; nptr = nptr->next)
    {
      if (nptr->flags & NS_COLLIDE)
      {
        struct Luser    *lptr;

        if ((lptr = FindClient(nptr->nick)))
        {
          if (unixtime >= nptr->collide_ts)
          {
            /*
             * The current time is 60+ seconds past the
             * last time they changed their nickname or
             * connected to the network, and they have not
             * yet identified - kill them
             */
            putlog(LOG1, 
              "%s: killing %s!%s@%s (Nickname Enforcement)",
              n_NickServ,
              lptr->nick,
              lptr->username,
              lptr->hostname);

            SendUmode(OPERUMODE_S,
              "%s: killing %s!%s@%s (Nickname Enforcement)",
              n_NickServ,
              lptr->nick,
              lptr->username,
              lptr->hostname);

            /*
             * kill the nick and replace with a pseudo nick
             */
            collide(lptr->nick);

            nptr->collide_ts = 0;
          }
        }
        else
        {
          /*
           * User must have changed their nick or QUIT -
           * remove the collide
           */
          nptr->flags &= ~NS_COLLIDE;
          nptr->collide_ts = 0;
        }
      } /* if (nptr->flags & NS_COLLIDE) */
      else if (NSReleaseTimeout && (nptr->flags & NS_RELEASE))
      {
        struct Luser *lptr;

        if ((lptr = FindClient(nptr->nick)))
        {
          if ((unixtime - lptr->nick_ts) >= NSReleaseTimeout)
          {
            putlog(LOG1,
              "%s: Releasing enforcement pseudo-nick [%s]",
              n_NickServ,
              lptr->nick);

            SendUmode(OPERUMODE_S,
              "%s: Releasing enforcement pseudo-nick [%s]",
              n_NickServ,
              lptr->nick);

            /* release the nickname */
            release(lptr->nick);
          }
        }
        else
        {
          /*
           * Something got messed up, kill the release flag
           */

#ifdef BNET
	  /* we shouldnt get here.. */
	  toserv(":%s QUIT :Released\n", nptr->nick);
#else
	  putlog(LOG1,
		 "%s: No client for enforced pseudo-nick [%s]!",
		 n_NickServ,
		 nptr->nick);
#endif

          nptr->flags &= ~NS_RELEASE;
#ifdef BNET
	  nptr->flags &= ~NS_SVSNICK;
#endif
        }
      }
    } /* for (nptr = nicklist[ii]; nptr; nptr = nptr->next) */
  } /* for (ii = 0; ii < NICKLIST_MAX; ii++) */
} /* CollisionCheck() */

#ifdef LINKED_NICKNAMES

/*
InsertLink()
 Insert the nick 'leaf' into 'hub's linked nickname list.
Return: 1 if successful,
        0 for NULL arguements,
        -1 if a circular link is detected,
        -2 if more than MaxLinks links are formed
*/

static int
InsertLink(struct NickInfo *hub, struct NickInfo *leaf)

{
  struct NickInfo *master, /* new master for link list */
                  *leafmaster, /* previous master for leaf's list */
                  *tmp;
  int lcnt; /* link count */


  if (!hub || !leaf)
    return (0);

  master = NULL;

  if (!hub->master)
  {
    /*
     * "hub" is either a plain nickname, not linked to anyone
     * else, or it is the master nickname for a particular
     * list. In either case, it will be the master nickname
     * for leaf and all of leaf's current links.
     */

    master = hub;
  }
  else
  {
    /*
     * "hub" is already part of a link list, since it has a
     * master. So, make sure leaf and all of leaf's current
     * links get their master entry set to hub's current
     * master
     */

    master = hub->master;
  }

  if (!master)
  {
    /* something is very wrong */
    return (0);
  }

  for (tmp = master, myi = 0; tmp && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
  {
    if (tmp == leaf)
    {
      /*
       * circular link: a user in the link is attempting to
       * link to another user in the same link
       */
      return (-1);
    }
  }

  if (master->numlinks)
    lcnt = master->numlinks;
  else
    lcnt = 1; /* master is a standalone nickname */

  if (leaf->master)
  {
    if (leaf->master->numlinks)
      lcnt += leaf->master->numlinks;
    else
      lcnt++;
  }
  else if (leaf->numlinks)
    lcnt += leaf->numlinks;
  else
    ++lcnt;

  if (MaxLinks && (lcnt > MaxLinks))
    return (-2);

  /*
   * Store leaf's current master into 'leafmaster'
   */
  if (leaf->master)
  {
    leafmaster = leaf->master;

    /*
     * leaf's current master will no longer be a list master -
     * set its master field to the new master
     */
    leaf->master->master = master;
  }
  else
    leafmaster = leaf;

  /*
   * Add 1 to link count to account for "leaf"
   */
  ++master->numlinks;

  if (!master->nextlink)
  {
    /*
     * master was previously a standalone nickname, add 1
     * to its link count to account for itself
     */
    ++master->numlinks;
  }

#ifdef CHANNELSERVICES

  /*
   * leafmaster should no longer have a FounderChannels list -
   * add all of leafmaster's channels to master's channels.
   * There's no point in reallocating - just assign master's
   * pointer to leafmaster's
   */
  master->FounderChannels = leafmaster->FounderChannels;
  master->fccnt = leafmaster->fccnt;
  leafmaster->FounderChannels = NULL;
  leafmaster->fccnt = 0;

#endif /* CHANNELSERVICES */

  /*
   * Now traverse leaf's link list and set all the master
   * entries to the "master" variable, which will point
   * to the rightful master of the new list
   */


  for (tmp = leafmaster, myi = 0; tmp->nextlink && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
  {
    tmp->master = master;
    ++master->numlinks;
  }

  /*
   * The above loop stopped without setting the very last
   * link's master, do that now.
   */
  tmp->master = master;

  /*
   * tmp now points to the very last nick structure in
   * leaf's link list. In order to keep leaf's current
   * link list intact, set tmp->nextlink to hub->nextlink,
   * and set hub->nextlink to leaf. The new list will
   * look something like:
   *   master -> hub -> leaflist -> hub->nextlink
   */

  tmp->nextlink = hub->nextlink;
  hub->nextlink = leafmaster;
  /*if (hub != master)
    master->nextlink = leafmaster;*/

  return (1);
} /* InsertLink() */

/*
DeleteLink()
 Remove nptr from it's current link list. If copyhosts == 1, copy nptr's
 master's access list

Return: 1  if successful
        0  if NULL pointer
        -1 if nptr is not in a link

XXX: We have bugs here. Fix them! -kre
*/

static int DeleteLink(struct NickInfo *nptr, int copyhosts)
{
  struct NickInfo *tmp, *master;
  struct NickHost *hptr;

  if (!nptr)
    return (0);

  if (!nptr->master && !nptr->nextlink)
    return (-1);

  if (!nptr->master && nptr->nextlink)
    /* "nptr" IS the master of the list */
    tmp = NULL;
  else
  {
    for (tmp = nptr->master, myi = 0; tmp && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
      if (tmp->nextlink == nptr)
        break;

    if (!tmp)
      return (0);
  }

  /*
   * "tmp" now points to the link structure right before nptr in the link.
   * If tmp is NULL, then nptr is the master of the list, and there is
   * no-one before it.
   */

  if (tmp)
  {
    master = nptr->master;
    tmp->nextlink = nptr->nextlink;
    nptr->master = nptr->nextlink = NULL;
    nptr->numlinks = 0;

    if (copyhosts)
    {
      /*
        * Now, since nptr is a standalone nick again, it needs
        * an access list - give it master's access list
        */
      for (hptr = master->hosts; hptr; hptr = hptr->next)
        AddHostToNick(hptr->hostmask, nptr);
    }
  }
  else
  {
    /*
     * The master (nptr) is being deleted, make nptr->nextlink the new
     * master - copy access list to new master as well.
     */

    nptr->nextlink->master = NULL;
    nptr->nextlink->numlinks = nptr->numlinks;

    /*
     * Go through list and set everyone's master entry to the new master
     * (nptr->nextlink)
     */
    for (tmp = nptr->nextlink->nextlink, myi = 0; tmp && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
      tmp->master = nptr->nextlink;

    if (copyhosts)
    {
      /*
       * Now give nptr->nextlink, nptr's access list
       */
      for (hptr = nptr->hosts; hptr; hptr = hptr->next)
        AddHostToNick(hptr->hostmask, nptr->nextlink);
    }

  #ifdef CHANNELSERVICES

    /*
     * The new master should keep the list of founder channels from the
     * old master
     */
    nptr->nextlink->FounderChannels = nptr->FounderChannels;
    nptr->nextlink->fccnt = nptr->fccnt;
    nptr->FounderChannels = NULL;
    nptr->fccnt = 0;

  #endif /* CHANNELSERVICES */

    master = nptr->nextlink;
  }

  if (!master)
    return (0); /* shouldn't happen */

  --master->numlinks;
  if (master->numlinks == 1)
  {
    /*
     * There were only 2 links in the list to begin with,
     * and we deleted one, so the master is no longer a
     * master
     */
    master->numlinks = 0;
    master->master = NULL;
  }

  return (1);
} /* DeleteLink() */

/*
IsLinked()
 Determine if nick1 and nick2 are in the same link.
Return 1 if yes, 0 if not
*/

int
IsLinked(struct NickInfo *nick1, struct NickInfo *nick2)

{
  struct NickInfo *tmp1, *tmp2;

  if (!nick1 || !nick2)
    return (0);

  /*
   * Check to see if nick1's master is equal to nick2's master -
   * if so, they are in the same link
   */

  if (nick1->master)
    tmp1 = nick1->master;
  else
    tmp1 = nick1;

  if (nick2->master)
    tmp2 = nick2->master;
  else
    tmp2 = nick2;

  if (tmp1 == tmp2)
    return (1);

  return (0);
} /* IsLinked() */

#endif /* LINKED_NICKNAMES */

/*
GetMaster()
 Return a pointer to nptr's nick link master
*/

struct NickInfo *
GetMaster(struct NickInfo *nptr)

{
  if (!nptr)
    return (NULL);

#ifdef LINKED_NICKNAMES
  if (nptr->master)
    return (nptr->master);
#endif

  return (nptr);
} /* GetMaster() */

static void
n_help(struct Luser *lptr, int ac, char **av)

{
  if (ac >= 2)
  {
    char  str[MAXLINE];

    if (ac >= 3)
      ircsprintf(str, "%s %s", av[1], av[2]);
    else
    {
      if ((!irccmp(av[1], "ACCESS")) ||
          (!irccmp(av[1], "SET")))
        ircsprintf(str, "%s index", av[1]);
      else
      {
        struct Command *cptr;

        for (cptr = nickcmds; cptr->cmd; cptr++)
          if (!irccmp(av[1], cptr->cmd))
            break;

        if (cptr->cmd)
          if ((cptr->level == LVL_ADMIN) &&
              !(IsValidAdmin(lptr)))
          {
            notice(n_NickServ, lptr->nick,
              "No help available on \002%s\002",
              av[1]);
            return;
          }

        ircsprintf(str, "%s", av[1]);
      }
    }

    GiveHelp(n_NickServ, lptr->nick, str, NODCC);
  }
  else
    GiveHelp(n_NickServ, lptr->nick, NULL, NODCC);
} /* n_help() */

/*
n_register()
  Add lptr->nick to nick database with password av[1]
*/

static void
n_register(struct Luser *lptr, int ac, char **av)

{
  char *mask;
  struct NickInfo *nptr;
  time_t currtime;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick, "Syntax: \002REGISTER\002 <password> <securepassword>");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "REGISTER");
    return;
  }

  if (ircncmp(lptr->nick, GuestNick? GuestNick : "Guest", 5) == 0)
    {
      notice(n_NickServ, lptr->nick,
	     "You cannot register nicks which begin with '%s'",
             GuestNick? GuestNick : "Guest");
      return;
    }

  if ((nptr = FindNick(lptr->nick)))
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] is already registered",
      nptr->nick);
    return;
  }

  currtime = current_ts;

  if (currtime < (lptr->nickreg_ts + NickRegDelay))
  {
    notice(n_NickServ, lptr->nick,
      "Please wait %ld seconds before using REGISTER again",
      NickRegDelay);
    return;
  }

  nptr = MakeNick();

  if (!ChangePass(nptr, av[1], av[2]))
  {
    notice(n_NickServ, lptr->nick,
      "Register failed");
    putlog(LOG1, "%s: failed to register %s!%s@%s",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);
    MyFree(nptr);
    return;
  }

  nptr->nick = MyStrdup(lptr->nick);
  nptr->created = nptr->lastseen = currtime;

  if (LastSeenInfo)
  {
    nptr->lastu = MyStrdup(lptr->username);
    nptr->lasth = MyStrdup(lptr->hostname);
    nptr->lastqmsg = NULL;
  }

  nptr->flags = NS_IDENTIFIED;

  if (NSSetKill)
    nptr->flags |= NS_PROTECTED;
  if (NSSetAutoMask)
    nptr->flags |= NS_AUTOMASK;
  if (NSSetPrivate)
    nptr->flags |= NS_PRIVATE;
  if (NSSetSecure)
    nptr->flags |= NS_SECURE;
  if (NSSetUnSecure)
    nptr->flags |= NS_UNSECURE;
  if (NSSetAllowMemos)
    nptr->flags |= NS_MEMOS;
  if (NSSetMemoSignon)
    nptr->flags |= NS_MEMOSIGNON;
  if (NSSetMemoNotify)
    nptr->flags |= NS_MEMONOTIFY;
  if (NSSetHide)
    nptr->flags |= NS_HIDEALL;
  if (NSSetHideEmail)
    nptr->flags |= NS_HIDEEMAIL;
  if (NSSetHideUrl)
    nptr->flags |= NS_HIDEURL;
  if (NSSetHideQuit)
    nptr->flags |= NS_HIDEQUIT;

  if (IsOperator(lptr))
    nptr->flags |= NS_OPERATOR;

  mask = HostToMask(lptr->username, lptr->hostname);

  AddHostToNick(mask, nptr);

  /* Add nptr to nick table */
  AddNick(nptr);

  lptr->nickreg_ts = currtime;

  notice(n_NickServ, lptr->nick,
    "Your nickname is now registered under the hostmask [\002%s\002]\n",
    mask);
  notice(n_NickServ, lptr->nick,
    "Your password is [\002%s\002] - Remember this for later use",
    av[1]);
  notice(n_NickServ, lptr->nick,
    "Your secure password is [\002%s\002] - You will need it to change the primary password or drop your nick",
    av[2]);

  MyFree(mask);

  RecordCommand("%s: %s!%s@%s REGISTER",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname);
} /* n_register() */

/*
n_drop()
  Discontinue registered nick 'lptr->nick'
*/

static void
n_drop(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *ni;
  char *dnick;
  struct Userlist *utmp;

  if (!(ni = FindNick(lptr->nick)))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      lptr->nick);
    return;
  }

  /*
   * Set dnick to lptr->nick instead of ni->nick because
   * when ni is deleted, dnick will be garbage
   */
  dnick = lptr->nick;

  if (lptr->flags & L_OSREGISTERED)
    utmp = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
  else
    utmp = GetUser(0, lptr->nick, lptr->username, lptr->hostname);

  /*  if (!(ni->flags & NS_IDENTIFIED))
   *
   * now we require the secure password to be supplied for DROP  -demond
   */
  if (!(IsAdmin(utmp) && IsValidAdmin(lptr)))
  {
    if (ac < 2)
    {
      notice(n_NickServ, lptr->nick, "Syntax: \002DROP\002 <securepassword>");
      notice(n_NickServ, lptr->nick,
        ERR_MORE_INFO,
        n_NickServ,
        "DROP");
      return;
    }
    else
    {
      if (!pwmatch(ni->securepass, av[1]))
      {
        notice(n_NickServ, lptr->nick, "Invalid secure password");
        RecordCommand("%s: %s!%s@%s failed DROP",
          n_NickServ,
          lptr->nick,
          lptr->username,
          lptr->hostname);
        return;
      }
    }
  }

  if (IsAdmin(utmp) && (ni->flags & NS_IDENTIFIED) && ac >= 2)
  {
#ifdef EMPOWERADMINS
      if (!IsValidAdmin(lptr))
      {
        notice(n_NickServ, lptr->nick,
          ERR_BAD_PASS);
        RecordCommand("%s: (Unregistered) Administrator %s!%s@%s failed DROP [%s]",
          n_NickServ,
          lptr->nick,
          lptr->username,
          lptr->hostname,
          av[1]);
        return;
      }

      if (!(ni = FindNick(av[1])))
      {
        notice(n_NickServ, lptr->nick,
          ERR_NOT_REGGED,
          av[1]);
        return;
      }

      dnick = av[1];

      RecordCommand("%s: Administrator %s!%s@%s DROP [%s]",
        n_NickServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        dnick);

       o_Wallops("DROP from %s!%s@%s for nick [%s]",
          lptr->nick, lptr->username, lptr->hostname,
          av[1] );
#endif
    }
  else
  {
    /* just a regular user dropping their nick */
    RecordCommand("%s: %s!%s@%s DROP",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);
  }

  /* remove the nick from the nicklist table */

  DeleteNick(ni);

  notice(n_NickServ, lptr->nick,
    "The nickname [\002%s\002] has been dropped",
    dnick);
} /* n_drop() */

/*
n_identify()
  If successful, lptr->nick is now recognized and can use other
NickServ commands
  If lptr->nick is in a nickname link, the identifying process
works something like this:
  Each individual nick in the link gets their own identified
flag when they identify.  If lptr->nick ever changes his/her nick
to another nickname in the same link, s_nick() will check if
lptr->nick was identified with NickServ.  If so, the new nick
will get auto-identified.  Otherwise the new nickname will have
to identify with NickServ manually. It might be easier and more
efficient to simply set the master of the link as identified
whenever a leaf identifies, but I have chosen not to do this because
a malicious user could wait until a leaf identifies, then change
their nickname to a nickname in the link, and become auto-identified.
Keeping separate identifying flags for each nickname in the link
will prevent this.
*/

static void
n_identify(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr, *realptr;
#ifdef MEMOSERVICES
  struct MemoInfo *mi;
#endif

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002IDENTIFY\002 <password>");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "IDENTIFY");
    return;
  }

  realptr = FindNick(lptr->nick);

  /*
   * Get master nickname as well for auto-mask purposes
   */
  nptr = GetMaster(realptr);

  if (!realptr || !nptr)
  {
    notice(n_NickServ, lptr->nick, ERR_NOT_REGGED,
      lptr->nick);
    return;
  }

  if (realptr->flags & NS_IDENTIFIED)
  {
    notice(n_NickServ, lptr->nick,
      "You have already identified");
    return;
  }

  if (!pwmatch(realptr->password, av[1]))
  {
    notice(n_NickServ, lptr->nick, ERR_BAD_PASS);
    RecordCommand("%s: %s!%s@%s failed IDENTIFY",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);
    return;
  }

  RecordCommand("%s: %s!%s@%s IDENTIFY",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname);

  realptr->flags |= NS_IDENTIFIED;
  realptr->flags &= ~(NS_COLLIDE | NS_RELEASE);
  notice(n_NickServ, lptr->nick,
    "Password accepted - you are now recognized");

  if ((nptr->flags & NS_AUTOMASK) &&
      (!OnAccessList(lptr->username, lptr->hostname, nptr)))
  {
    /*
     * well, i am tired of all those dynamic ident Java lamurs... 
     *                                            -demond
     */
#ifndef BNET 
    char  *mask = HostToMask(lptr->username, lptr->hostname);
#else
    char  *mask = HostToMask("", lptr->hostname);
#endif
    int added = AddHostToNick(mask, nptr);

    if (added) {
      notice(n_NickServ, lptr->nick,
	"Added hostmask [\002%s\002] to the list of your known hosts", mask);
    } else {
      notice(n_NickServ, lptr->nick,
	"Your access list is full (%d hostmasks). Automask was NOT performed", MaxNickHosts);
      notice(n_NickServ, lptr->nick,
	"Please use the ACCESS DEL command to purge unnecessary entries");
    }

    MyFree(mask);
  }

  if (IsOperator(lptr))
    CheckOper(lptr->nick);

  if (LastSeenInfo)
  {
    /*
     * Update last seen user@host info
     */

    if (realptr->lastu)
      MyFree(realptr->lastu);
    if (realptr->lasth)
      MyFree(realptr->lasth);
    realptr->lastu = MyStrdup(lptr->username);
    realptr->lasth = MyStrdup(lptr->hostname);
  } /* if (LastSeenInfo) */

  /* I have decided to move here new memo checking code, because it
   * seems to me more reasonable to have it right after successful
   * identify, and not every time on signon -kre */
#ifdef MEMOSERVICES
  if (nptr->flags & NS_MEMOSIGNON)
  {
    if ((mi = FindMemoList(lptr->nick)))
    {
      if (mi->newmemos)
      {
        notice(n_MemoServ, lptr->nick,
          "You have \002%d\002 new memo%s",
          mi->newmemos,
          (mi->newmemos == 1) ? "" : "s");
        notice(n_MemoServ, lptr->nick,
          "Type \002/msg %s LIST\002 to view them",
          n_MemoServ);
      }
    }
    else
      notice(n_MemoServ, lptr->nick,
          "You have no new memos");
  }
#endif /* MEMOSERVICES */

#ifdef BNET
  toserv(":%s SVSHINT 1 :%s\n", Me.name, lptr->nick);
#endif
} /* n_identify() */

/*
n_recover()
  Recover nickname av[1] through a collide, and hold it with a
pseudo-nick
*/

static void
n_recover(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *ni;
  int goodcoll;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002RECOVER <nickname> [password]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "RECOVER");
    return;
  }

  if (!(ni = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  /*
   * Check if lptr's userhost is on ni's access list - if so check
   * if ni is SECURE (if it is, lptr->nick MUST supply the correct
   * password)
   */
  goodcoll = 0;
  if (OnAccessList(lptr->username, lptr->hostname, ni)
      && (!(ni->flags & NS_SECURE)))
    goodcoll = 1;
  else
  {
    if (ac >= 3)
      if (pwmatch(ni->password, av[2]))
        goodcoll = 1;
  }

  if (!goodcoll)
  {
    /* Admins may recover any nickname */
    if (IsValidAdmin(lptr))
      goodcoll = 2;
  }

  if (goodcoll)
  {
    if (ni->flags & NS_RELEASE)
    {
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] has already been recovered",
        av[1]);
      return;
    }

    if (!FindClient(av[1]))
    {
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] is not currently online",
        av[1]);
      return;
    }

    collide(av[1]);
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] has been recovered",
      av[1]);
    notice(n_NickServ, lptr->nick,
      "Type: \002/msg %s RELEASE %s\002 to release the "
      "nickname before the timeout",
        n_NickServ,
        av[1]);

    RecordCommand("%s: %s%s!%s@%s RECOVER [%s]",
      n_NickServ,
      (goodcoll == 2) ? "Administrator " : "",
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
  else
  {
    notice(n_NickServ, lptr->nick,
      ERR_BAD_PASS);
    RecordCommand("%s: %s!%s@%s failed RECOVER [%s]",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
} /* n_recover() */

/*
n_release()
  Release a pseudo nick that is being held after a RECOVER
*/

static void
n_release(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *ni;
  int goodrel;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002RELEASE <nickname> [password]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "RELEASE");
    return;
  }

  if (!(ni = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  /*
   * Check if lptr's userhost is on ni's access list - if so check
   * if ni is SECURE (if it is, lptr->nick MUST supply the correct
   * password)
   */
  goodrel = 0;
  if (OnAccessList(lptr->username, lptr->hostname, ni) &&
      (!(ni->flags & NS_SECURE)))
    goodrel = 1;
  else
  {
    if (ac >= 3)
      if (pwmatch(ni->password, av[2]))
        goodrel = 1;
  }

  if (!goodrel)
  {
    /* Admins who are IDENTIFY'd may release any nickname */
    if (IsValidAdmin(lptr))
      goodrel = 2;
  }

  if (goodrel)
  {
    if (!(ni->flags & NS_RELEASE))
    {
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] is not being enforced",
        av[1]);
      return;
    }

    if (!FindClient(av[1]))
    {
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] is not currently online",
        av[1]);
      return;
    }

    release(av[1]);
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] has been released",
      av[1]);

    RecordCommand("%s: %s%s!%s@%s RELEASE [%s]",
      n_NickServ,
      (goodrel == 2) ? "Administrator " : "",
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
  else
  {
    notice(n_NickServ, lptr->nick,
      ERR_BAD_PASS);

    RecordCommand("%s: %s!%s@%s failed RELEASE [%s]",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
} /* n_release() */

/*
n_ghost()
  Kill nickname av[1] if av[2] matches lptr's password. (for clients
who have ghosted nicknames)
*/

static void
n_ghost(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *ni;
  int goodcoll;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002GHOST <nickname> [password]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "GHOST");
    return;
  }

  if (!(ni = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  /*
   * Check if lptr's userhost is on ni's access list - if so check
   * if ni is SECURE (if it is, lptr->nick MUST supply the correct
   * password)
   */
  goodcoll = 0;
  if (OnAccessList(lptr->username, lptr->hostname, ni) &&
      (!(ni->flags & NS_SECURE)))
    goodcoll = 1;
  else
  {
    if (ac >= 3)
      if (pwmatch(ni->password, av[2]))
        goodcoll = 1;
  }

  if (!goodcoll)
  {
    /* Admins may kill any nickname */
    if (IsValidAdmin(lptr))
      {
        goodcoll = 1;
        o_Wallops("Administrative ghost from %s for nick [%s]",
                  lptr->nick, av[1]);
      }
  }

  if (goodcoll)
  {
    struct Luser    *gptr;

    if (!(gptr = FindClient(av[1])))
    {
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] is not currently online",
        av[1]);
      return;
    }

    toserv(":%s KILL %s :%s!%s (Ghost: %s!%s@%s)\n",
      n_NickServ,
      av[1],
      Me.name,
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

    DeleteClient(gptr);

    notice(n_NickServ, lptr->nick,
      "[\002%s\002] has been killed",
      av[1]);

    RecordCommand("%s: %s!%s@%s GHOST [%s]",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
  else
  {
    notice(n_NickServ, lptr->nick,
      ERR_BAD_PASS);
    RecordCommand("%s: %s!%s@%s failed GHOST [%s]",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
  }
} /* n_ghost() */

/*
n_access()
  Modify the list of known hostmasks for lptr->nick
*/

static void
n_access(struct Luser *lptr, int ac, char **av)

{
  struct Command *cptr;
  struct NickInfo *target;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002ACCESS {ADD|DEL|LIST} [mask]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "ACCESS");
    return;
  }

  target = NULL;

#ifdef EMPOWERADMINS_MORE

  /*
   * Allow administrators to specify a nickname to modify
   */
  if (IsValidServicesAdmin(lptr))
  {
    /*
     * First, check if av[1] is a valid command. If not,
     * check if it is a valid nickname. If av[1] is
     * a valid command, process the command as though
     * lptr is modifying their own access, which they
     * most likely are. If av[1] is neither a valid
     * command nor a valid nickname, give them
     * a "Nick not registered" error.
     */
    cptr = GetCommand(accesscmds, av[1]);
    if (!cptr || (cptr == (struct Command *) -1))
    {
      target = GetLink(av[1]);
      if (!target)
      {
        notice(n_NickServ, lptr->nick,
          ERR_NOT_REGGED,
          av[1]);
        return;
      }

      if (ac >= 3)
        cptr = GetCommand(accesscmds, av[2]);
      else
        cptr = NULL;
    }
  }
  else
    cptr = GetCommand(accesscmds, av[1]);

#else

  cptr = GetCommand(accesscmds, av[1]);

#endif /* EMPOWERADMINS_MORE */

  if (cptr && (cptr != (struct Command *) -1))
  {
    /* call the appropriate function */
    (*cptr->func)(lptr, target, ac, av);
  }
  else
  {
    /* the command doesn't exist */
    notice(n_NickServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[1]);
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "ACCESS");
  }
} /* n_access() */

static void
n_access_add(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  struct NickInfo *nptr;
  char *mask,
       *host,
       *tmp;

  char user[USERLEN];
  
  memset( (char*)user, 0x00, USERLEN );

  if (target)
  {
    if (ac < 4)
    {
      notice(n_NickServ, lptr->nick,
        "Syntax: \002ACCESS <nickname> ADD <mask>\002");
      notice(n_NickServ, lptr->nick,
        ERR_MORE_INFO, 
        n_NickServ,
        "ACCESS ADD");
      return;
    }

    nptr = target;
    mask = av[3];
  }
  else
  {
    if (ac < 3)
    {
      notice(n_NickServ, lptr->nick, "Syntax: \002ACCESS ADD <mask>\002");
      notice(n_NickServ, lptr->nick,
        ERR_MORE_INFO, 
        n_NickServ,
        "ACCESS ADD");
      return;
    }

    nptr = GetLink(lptr->nick);
    mask = av[2];
  }

  if(strlen(mask) > UHOSTLEN ) {
	notice(n_NickServ, lptr->nick, "Hostmask too long!");
        return;
  }

  if (!nptr)
    return;

  tmp = MyStrdup(mask);
  
  host = strchr(tmp, '@');
  if( host == NULL ) {
        notice(n_NickServ, lptr->nick, "Invalid hostmask!");
        return;
  }
  if( strlen(mask)-strlen(host) > USERLEN - 1 ) {
        notice(n_NickServ, lptr->nick, "Invalid hostmask!  Username too long!");
        return;
  }

  if(strlen(host)) {
    strncpy(user,mask,strlen(mask)-strlen(host));
  }

  if (!user || !host)
  {
    notice(n_NickServ, lptr->nick,
      "The hostmask [\002%s\002] is invalid",
      mask);
    return;
  }

  ++host;

  if (OnAccessList(user, host+1, nptr))
  {
    if (target)
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] matches a hostmask already on the Access List for [%s]",
          mask,
          target->nick);
    else
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] matches a hostmask already on your Access List",
        mask);
    return;
  }
  else
  {
    int added = AddHostToNick(mask, nptr);

    if (target)
    {
      if (added) {
	notice(n_NickServ, lptr->nick,
	       "[\002%s\002] has been added to the Access List for [%s]",
	       mask,
	       target->nick);

	RecordCommand("%s: Administrator %s!%s@%s ACCESS %s ADD [%s]",
		      n_NickServ,
		      lptr->nick,
		      lptr->username,
		      lptr->hostname,
		      target->nick,
		      mask);
      } else {
	notice(n_NickServ, lptr->nick,
	       "Access list for [\002%s\002] is full (%d hostmasks). Hostmask NOT added", target->nick, MaxNickHosts);
      }
    }
    else
    {
      if (added) {
	notice(n_NickServ, lptr->nick,
	       "[\002%s\002] added to your Access List",
	       mask);

	RecordCommand("%s: %s!%s@%s ACCESS ADD [%s]",
		      n_NickServ,
		      lptr->nick,
		      lptr->username,
		      lptr->hostname,
		      mask);
      } else {
	notice(n_NickServ, lptr->nick,
	       "Your access list is full (%d hostmasks). The specified hostmask was NOT added", MaxNickHosts);
        notice(n_NickServ, lptr->nick,
               "Please use the ACCESS DEL command to purge unnecessary entries");
      }
    }
  }
} /* n_access_add() */

static void
n_access_del(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  struct NickInfo *nptr;
  struct NickHost *hptr, *prev;
  int found;
  char *mask;

  if (target)
  {
    if (ac < 4)
    {
      notice(n_NickServ, lptr->nick,
        "Syntax: \002ACCESS <nickname> DEL <mask>\002");
      notice(n_NickServ, lptr->nick,
        ERR_MORE_INFO, 
        n_NickServ,
        "ACCESS DEL");
      return;
    }

    nptr = target;
    mask = av[3];
  }
  else
  {
    if (ac < 3)
    {
      notice(n_NickServ, lptr->nick,
        "Syntax: \002ACCESS DEL <mask>\002");
      notice(n_NickServ, lptr->nick,
        ERR_MORE_INFO, 
        n_NickServ,
        "ACCESS DEL");
      return;
    }

    nptr = GetLink(lptr->nick);
    mask = av[2];
  }

  if (!nptr)
    return;

  found = 0;

  prev = NULL;
  for (hptr = nptr->hosts; hptr; )
  {
    if (match(mask, hptr->hostmask))
    {
      found = 1;
      MyFree(hptr->hostmask);

      if (prev)
      {
        prev->next = hptr->next;
        MyFree(hptr);
        hptr = prev;
      }
      else
      {
        nptr->hosts = hptr->next;
        MyFree(hptr);
        hptr = NULL;
      }

      break;
    }

    prev = hptr;

    if (hptr)
      hptr = hptr->next;
    else
      hptr = nptr->hosts;
  }

  if (found)
  {
    if (target)
    {
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] removed from the Access List for [%s]",
        mask,
        target->nick);

      RecordCommand("%s: Administrator %s!%s@%s ACCESS %s DEL [%s]",
        n_NickServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        target->nick,
        mask);
    }
    else
    {
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] removed from your Access List",
        mask);

      RecordCommand("%s: %s!%s@%s ACCESS DEL [%s]",
        n_NickServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        mask);
    }
  }
  else
  {
    if (target)
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] was not found on the Acces List for [%s]",
        mask,
        target->nick);
    else
      notice(n_NickServ, lptr->nick,
        "[\002%s\002] was not found on your Access List",
        mask);
  }
} /* n_access_del() */

static void
n_access_list(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  struct NickInfo *nptr;
  int cnt = 1;
  struct NickHost *hptr;
  char *mask;

  if (target)
    nptr = target;
  else
    nptr = GetLink(lptr->nick);

  if (!nptr)
    return;

  mask = NULL;

  if (target)
  {
    if (ac >= 4)
      mask = av[3];
  }
  else
  {
    if (ac >= 3)
      mask = av[2];
  }

  if (target)
    RecordCommand("%s: Administrator %s!%s@%s ACCESS %s LIST %s",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      target->nick,
      mask ? mask : "");
  else
    RecordCommand("%s: %s!%s@%s ACCESS LIST %s",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      mask ? mask : "");

  notice(n_NickServ, lptr->nick,
    "-- Access List for [\002%s\002] --",
    nptr->nick);

  for (hptr = nptr->hosts; hptr; hptr = hptr->next)
  {
    if (mask)
      if (match(mask, hptr->hostmask) == 0)
        continue;

    notice(n_NickServ, lptr->nick, "%d) %s",
      cnt++,
      hptr->hostmask);
  }

  notice(n_NickServ, lptr->nick, "-- End of list --");
} /* n_access_list() */

/*
n_set()
  Set various options for your nick
*/

static void
n_set(struct Luser *lptr, int ac, char **av)

{
  struct Command *cptr;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET <option> [parameter]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "SET");
    return;
  }

  cptr = GetCommand(setcmds, av[1]);

  if (cptr && (cptr != (struct Command *) -1))
  {
    /* call the appropriate function */
    (*cptr->func)(lptr, ac, av);
  }
  else
  {
    /* the command doesn't exist */
    notice(n_NickServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[1]);
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "SET");
    return;
  }
} /* n_set() */

static void
n_set_kill(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET KILL %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Kill Protection for your nickname is [\002%s\002]",
      (nptr->flags & NS_PROTECTED) ? "ON" : ((nptr->flags & NS_KILLIMMED) ? "IMMED" : "OFF"));
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags &= ~NS_KILLIMMED;
    if (AllowKillProtection)
    {
      nptr->flags |= NS_PROTECTED;
      notice(n_NickServ, lptr->nick,
        "Toggled Kill Protection [\002ON\002]");
      return;
    }
    notice(n_NickServ, lptr->nick,
      "Kill Protection is disabled on this network");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~(NS_PROTECTED | NS_KILLIMMED);
    notice(n_NickServ, lptr->nick,
      "Toggled Kill Protection [\002OFF\002]");
    return;
  }

  if (!irccmp(av[2], "IMMED"))
  {
    if (AllowKillImmed)
    {
      nptr->flags &= ~NS_PROTECTED;
      nptr->flags |= NS_KILLIMMED;
      notice(n_NickServ, lptr->nick,
        "Toggled Immediate Kill Protection [\002ON\002]");
    }
    else
    {
      notice(n_NickServ, lptr->nick,
        "The [\002IMMED\002] option is disabled");
    }

    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET KILL {ON|OFF|IMMED}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET KILL");
} /* n_set_kill() */

static void
n_set_automask(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET AUTOMASK %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "AutoMask for your nickname is [\002%s\002]",
      (nptr->flags & NS_AUTOMASK) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_AUTOMASK;
    notice(n_NickServ, lptr->nick,
      "Toggled AutoMask [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_AUTOMASK;
    notice(n_NickServ, lptr->nick,
      "Toggled AutoMask [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET AUTOMASK {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET AUTOMASK");
} /* n_set_automask() */

static void
n_set_private(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET PRIVATE %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Privacy for your nickname is [\002%s\002]",
      (nptr->flags & NS_PRIVATE) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_PRIVATE;
    notice(n_NickServ, lptr->nick,
      "Toggled Privacy [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_PRIVATE;
    notice(n_NickServ, lptr->nick,
      "Toggled Privacy [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET PRIVATE {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET PRIVATE");
} /* n_set_private() */

static void
n_set_oper(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  if (ac < 3)
  {
    RecordCommand("%s: %s!%s@%s SET OPER",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

    notice(n_NickServ, lptr->nick,
      "Operator Status for your nickname is [\002%s\002]",
      (nptr->flags & NS_OPERATOR) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    if (!IsOperator(lptr))
    {
      notice(n_NickServ, lptr->nick,
        "Permission Denied - You are not an IRC Operator");

      RecordCommand("%s: %s!%s@%s failed attempt to use SET OPER ON",
        n_NickServ,
        lptr->nick,
        lptr->username,
        lptr->hostname);

      return;
    }

    RecordCommand("%s: %s!%s@%s SET OPER ON",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

    nptr->flags |= NS_OPERATOR;
    notice(n_NickServ, lptr->nick,
      "Toggled Operator Status [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    RecordCommand("%s: %s!%s@%s SET OPER OFF",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

    nptr->flags &= ~NS_OPERATOR;
    notice(n_NickServ, lptr->nick,
      "Toggled Operator Status [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET OPER {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET OPER");
} /* n_set_oper() */

static void
n_set_secure(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET SECURE %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Security for your nickname is [\002%s\002]",
      (nptr->flags & NS_SECURE) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_SECURE;
    notice(n_NickServ, lptr->nick,
      "Toggled Security [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_SECURE;
    notice(n_NickServ, lptr->nick,
      "Toggled Security [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET SECURE {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET SECURE");
} /* n_set_secure() */

static void
n_set_unsecure(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET UNSECURE %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "UnSecure for your nickname is [\002%s\002]",
      (nptr->flags & NS_UNSECURE) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_UNSECURE;
    notice(n_NickServ, lptr->nick,
      "Toggled UnSecure [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_UNSECURE;
    notice(n_NickServ, lptr->nick,
      "Toggled UnSecure [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET UNSECURE {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET UNSECURE");
} /* n_set_unsecure() */

static void
n_set_memos(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET MEMOS %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Allow Memos for your nickname is [\002%s\002]",
      (nptr->flags & NS_MEMOS) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_MEMOS;
    notice(n_NickServ, lptr->nick,
      "Toggled Allow Memos [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_MEMOS;
    notice(n_NickServ, lptr->nick,
      "Toggled Allow Memos [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET MEMOS {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET MEMOS");
} /* n_set_memos() */

static void
n_set_notify(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET NOTIFY %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Notification of New Memos for your nickname is [\002%s\002]",
      (nptr->flags & NS_MEMONOTIFY) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_MEMONOTIFY;
    notice(n_NickServ, lptr->nick,
      "Toggled Notification of New Memos [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_MEMONOTIFY;
    notice(n_NickServ, lptr->nick,
      "Toggled Notification of New Memos [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET NOTIFY {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET NOTIFY");
} /* n_set_notify() */

static void
n_set_helplang(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET HELPLANG %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Help language for your nickname is [\002%s\002]",
      (nptr->flags & NS_BGHELP) ? "Bulgarian" : "English");
    return;
  }

  if (!irccmp(av[2], "BUL"))
  {
    nptr->flags |= NS_BGHELP;
    notice(n_NickServ, lptr->nick,
      "Help language set to [\002Bulgarian\002]");
    return;
  }

  if (!irccmp(av[2], "ENG"))
  {
    nptr->flags &= ~NS_BGHELP;
    notice(n_NickServ, lptr->nick,
      "Help language set to [\002English\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET HELPLANG {BUL|ENG}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET HELPLANG");
} /* n_set_helplang() */

static void
n_set_signon(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET SIGNON %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    (ac < 3) ? "" : StrToupper(av[2]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Notification of Memos on Signon for your nickname is [\002%s\002]",
      (nptr->flags & NS_MEMOSIGNON) ? "ON" : "OFF");
    return;
  }

  if (!irccmp(av[2], "ON"))
  {
    nptr->flags |= NS_MEMOSIGNON;
    notice(n_NickServ, lptr->nick,
      "Toggled Notification of Memos on Signon [\002ON\002]");
    return;
  }

  if (!irccmp(av[2], "OFF"))
  {
    nptr->flags &= ~NS_MEMOSIGNON;
    notice(n_NickServ, lptr->nick,
      "Toggled Notification of Memos on Signon [\002OFF\002]");
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET SIGNON {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET SIGNON");
} /* n_set_signon() */

static void
n_set_hide(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;
  char str[MAXLINE];
  int flag;

  if (!(nptr = GetLink(lptr->nick)))
    return;

  if (ac >= 3)
    strcpy(str, StrToupper(av[2]));

  RecordCommand("%s: %s!%s@%s SET HIDE %s %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    str, /* can't have 2 StrToupper()'s in the same stmt (static) */
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET HIDE {ALL|EMAIL|URL|QUIT} {ON|OFF}\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SET HIDE");
    return;
  }

  flag = 0;
  if (!ircncmp(av[2], "ALL", strlen(av[2])))
  {
    flag = NS_HIDEALL;
    strcpy(str, "Hide Info");
  }
  else if (!ircncmp(av[2], "EMAIL", strlen(av[2])))
  {
    flag = NS_HIDEEMAIL;
    strcpy(str, "Hide Email");
  }
  else if (!ircncmp(av[2], "URL", strlen(av[2])))
  {
    flag = NS_HIDEURL;
    strcpy(str, "Hide Url");
  }
  else if (!ircncmp(av[2], "QUIT", strlen(av[2])))
  {
    flag = NS_HIDEQUIT;
    strcpy(str, "Hide Quit");
  }
  else if (!ircncmp(av[2], "ADDR", strlen(av[2])))
  {
    flag = NS_HIDEADDR;
    strcpy(str, "Hide Address");
  }

  if (ac < 4)
  {
    notice(n_NickServ, lptr->nick,
      "%s for your nickname is [\002%s\002]",
      str,
      (nptr->flags & flag) ? "ON" : "OFF");
    return;
  }

  if (!flag)
  {
    notice(n_NickServ, lptr->nick,
      "Unknown switch [%s]",
      av[2]);
    return;
  }

  if (!irccmp(av[3], "ON"))
  {
    nptr->flags |= flag;
    notice(n_NickServ, lptr->nick,
      "Toggled %s [\002ON\002]",
      str);
    return;
  }

  if (!irccmp(av[3], "OFF"))
  {
    nptr->flags &= ~flag;
    notice(n_NickServ, lptr->nick,
      "Toggled %s [\002OFF\002]",
      str);
    return;
  }

  /* user gave an unknown param */
  notice(n_NickServ, lptr->nick,
    "Syntax: \002SET HIDE {ALL|EMAIL|URL|QUIT} {ON|OFF}\002");
  notice(n_NickServ, lptr->nick,
    ERR_MORE_INFO,
    n_NickServ,
    "SET HIDE");
} /* n_set_hide() */

static void
n_set_password(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 4)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET PASSWORD <secure/oldpass> <newpass>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SET PASSWORD");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  if (!pwmatch(nptr->securepass, av[2]))
  {
    notice(n_NickServ, lptr->nick, "Invalid secure password");
    return;
  }

  RecordCommand("%s: %s!%s@%s SET PASSWORD ...",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname);

  if (!ChangePass(nptr, av[3], av[2]))
  {
    notice(n_NickServ, lptr->nick,
      "Password change failed");
    return;
  }

  notice(n_NickServ, lptr->nick,
    "Your password has been changed to [\002%s\002]",
    av[3]);
} /* n_set_password() */

static void
n_set_email(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET EMAIL <email address|->\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SET EMAIL");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET EMAIL %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[2]);

  if (nptr->email)
    MyFree(nptr->email);

  if (!irccmp(av[2], "-"))
  {
    nptr->email = NULL;
    notice(n_NickServ, lptr->nick,
      "Your email address has been cleared");
    return;
  }

  nptr->email = MyStrdup(av[2]);

  notice(n_NickServ, lptr->nick,
    "Your email address has been set to [\002%s\002]",
    nptr->email);
} /* n_set_email() */

static void
n_set_url(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET URL <url|->\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SET URL");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET URL %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[2]);

  if (nptr->url)
    MyFree(nptr->url);

  if (!irccmp(av[2], "-"))
  {
    nptr->url = NULL;
    notice(n_NickServ, lptr->nick,
      "Your url has been cleared");
    return;
  }

  nptr->url = MyStrdup(av[2]);

  notice(n_NickServ, lptr->nick,
    "Your url has been set to [\002%s\002]",
    nptr->url);
} /* n_set_url() */

/* Set UIN for nickname. Taken from IrcBg and modified -kre */
static void n_set_uin(struct Luser *lptr, int ac, char **av)
{
  struct NickInfo *nptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick, "Syntax: \002SET UIN <uin|->\002");
    notice(n_NickServ, lptr->nick, ERR_MORE_INFO, n_NickServ, "SET UIN");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET UIN %s",
    n_NickServ, lptr->nick, lptr->username, lptr->hostname, av[2]);

  if (nptr->UIN)
    MyFree(nptr->UIN);

  if (!irccmp(av[2], "-"))
  {
    nptr->UIN = NULL;
    notice(n_NickServ, lptr->nick,
      "Your UIN has been cleared");
    return;
  }

  /* KrisDuv's check for validity of UIN */
  if (!IsNum(av[2]))
  {
    notice(n_NickServ, lptr->nick,
      "Invalid UIN [\002%s\002]", av[2]);
    return;
  }
  
  nptr->UIN = MyStrdup(av[2]);

  notice(n_NickServ, lptr->nick,
    "Your UIN has been set to [\002%s\002]", nptr->UIN);

} /* n_set_uin() */

/* Set GSM number for nickname. Taken from IrcBg and modified -kre */
static void n_set_gsm(struct Luser *lptr, int ac, char **av)
{
  struct NickInfo *nptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET GSM <GSM number|->\002");
    notice(n_NickServ, lptr->nick, ERR_MORE_INFO, n_NickServ,
      "SET GSM");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET GSM %s", n_NickServ, lptr->nick,
      lptr->username, lptr->hostname, av[2]);

  if (nptr->gsm)
    MyFree(nptr->gsm);

  if (!irccmp(av[2], "-"))
  {
    nptr->gsm = NULL;
    notice(n_NickServ, lptr->nick, "Your GSM number has been cleared.");
    return;
  }

  nptr->gsm = MyStrdup(av[2]);

  notice(n_NickServ, lptr->nick,
    "Your GSM number has been set to [\002%s\002]",
    nptr->gsm);
} /* n_set_gsm() */

/* Set phone number for nick. Code taken from IrcBg and modified -kre */
static void n_set_phone(struct Luser *lptr, int ac, char **av)
{
  struct NickInfo *nptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET PHONE <phone number|->\002");
    notice(n_NickServ, lptr->nick, ERR_MORE_INFO, n_NickServ,
      "SET PHONE");
    return;
  }

  if (!(nptr = FindNick(lptr->nick)))
    return;

  RecordCommand("%s: %s!%s@%s SET PHONE %s",
    n_NickServ, lptr->nick, lptr->username, lptr->hostname, av[2]);

  if (nptr->phone)
    MyFree(nptr->phone);

  if (!irccmp(av[2], "-"))
  {
    nptr->phone = NULL;
    notice(n_NickServ, lptr->nick, "Your phone number has been cleared");
    return;
  }

  nptr->phone = MyStrdup(av[2]);

  notice(n_NickServ, lptr->nick,
    "Your phone number has been set to [\002%s\002]",
    nptr->phone);

} /* n_set_phone() */

/* Clear all noexpire flags for all users. Code taken from IrcBg and
 * modified. -kre */
void n_clearnoexp(struct Luser *lptr, int ac, char **av)
{
  int ii, all = 1;
  struct NickInfo *nptr;
  long timeago = 0;

  if (!IsValidServicesAdmin(lptr))
    {
      notice(n_NickServ, lptr->nick, "You need to identify as services admin (+s) to be able to use this");
      return;
    }

  if (ac > 1)
    {
      all = 0;
      timeago = timestr(av[1]);
      if (timeago == 0) timeago = -1;
    }

  RecordCommand("%s: %s!%s@%s CLEARNOEXP %s",
    n_NickServ, lptr->nick, lptr->username, lptr->hostname,
    ac>1? av[1] : "");

  for (ii = 0; ii < NICKLIST_MAX; ++ii)
    for (nptr = nicklist[ii]; nptr; nptr = nptr->next)
      if (all || (!all && (nptr->lastseen < (current_ts - timeago))))
	{
	  nptr->flags &= ~NS_OPERATOR;
	  nptr->lastseen = current_ts; /* give'em a break... */
	  if (!all)
	    notice(n_NickServ, lptr->nick, "Cleared: %s", nptr->nick);
	}

  if (all)
    notice(n_NickServ, lptr->nick,
      "All noexpire flags for nicks have been cleared.");

} /* n_clearnoexp() */

#ifdef LINKED_NICKNAMES

/*
n_set_master()
 Set the master nickname for lptr's nick link to av[1]
*/

static void
n_set_master(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr,
                  *newmaster,
                  *origmaster;
  struct NickHost *hptr;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SET MASTER <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SET MASTER");
    return;
  }

  RecordCommand("%s: %s!%s@%s SET MASTER %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[2]);

  if (!(nptr = FindNick(lptr->nick)))
    return;

  if (!(newmaster = FindNick(av[2])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[2]);
    return;
  }

  if (!IsLinked(newmaster, nptr))
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] is not in your nickname link",
      newmaster->nick);
    return;
  }

  if ((nptr->master == newmaster) ||
      (!nptr->master && (nptr == newmaster)))
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] is already the master for your nickname link",
      newmaster->nick);
    return;
  }

  if (nptr->master)
    origmaster = nptr->master;
  else
    origmaster = nptr;

  DeleteLink(newmaster, 1);
  InsertLink(newmaster, origmaster);

  /*
   * Delete origmaster's access list, since newmaster
   * has it now
   */
  while (origmaster->hosts)
  {
    hptr = origmaster->hosts->next;
    MyFree(origmaster->hosts->hostmask);
    MyFree(origmaster->hosts);
    origmaster->hosts = hptr;
  }

  notice(n_NickServ, lptr->nick,
    "The master nickname for your link has been changed to [\002%s\002]",
    newmaster->nick);
} /* n_set_master() */

#endif /* LINKED_NICKNAMES */

/*
n_list()
  Display list of nicknames matching av[1]
*/

static void
n_list(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *temp;
  int IsAnAdmin;
  int ii,
      mcnt, /* total matches found */
      acnt; /* total matches - private nicks */
  enum {forbid, noexpire, online, private} opt = -1;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002LIST <pattern> [options]\002");
    notice(n_NickServ, lptr->nick, ERR_MORE_INFO, 
      n_NickServ, "LIST");
    return;
  }

  if (ac > 2)
  {
    if (!irccmp(av[2], "-b")) opt = forbid;
    else if (!irccmp(av[2], "-n")) opt = noexpire;
    else if (!irccmp(av[2], "-o")) opt = online;
    else if (!irccmp(av[2], "-p")) opt = private;
    else 
    {
      notice(n_NickServ, lptr->nick, "Invalid LIST option, must be [-b|-n|-o|-p]");      
      return;
    }
  }

  RecordCommand("%s: %s!%s@%s LIST %s %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1], ac>2? av[2] : "");

  if (IsValidAdmin(lptr))
    IsAnAdmin = 1;
  else
    IsAnAdmin = 0;

  acnt = mcnt = 0;
  notice(n_NickServ, lptr->nick,
    "-- Listing nicknames matching [\002%s\002] --",
    av[1]);
  for (ii = 0; ii < NICKLIST_MAX; ++ii)
  {
    for (temp = nicklist[ii]; temp; temp = temp->next)
    {
      if (match(av[1], temp->nick) && (IsAnAdmin? 1 : (mcnt < 255)))
      {
        ++mcnt;

        if ((IsAnAdmin) || !(temp->flags & NS_PRIVATE))
        {
          char str[20];

          if (opt==forbid && !(temp->flags & NS_FORBID)) continue;
	  if (opt==noexpire && !(temp->flags & NS_OPERATOR)) continue;
          else if (opt==online && !((temp->flags & NS_IDENTIFIED) && FindClient(temp->nick))) continue;
          else if (opt==private && !(temp->flags & NS_PRIVATE)) continue;

          ++acnt;
          if (temp->flags & NS_FORBID)
            strcpy(str, "<< FORBIDDEN >>");
          if (temp->flags & NS_OPERATOR)
            strcpy(str, "<< NOEXPIRE >>");	  
          else if ((temp->flags & NS_IDENTIFIED) && FindClient(temp->nick))
            strcpy(str, "<< ONLINE >>");
          else if (temp->flags & NS_PRIVATE)
            strcpy(str, "<< PRIVATE >>");
          else
            str[0] = '\0';

          notice(n_NickServ, lptr->nick, 
            "%-15s %15s %s %s ago",
            temp->nick,
            str,
	    opt==noexpire? "last seen" : "created",
            timeago(opt==noexpire? temp->lastseen : temp->created, 1));
        }
      } /* if (match(av[1], temp->nick)) */
    } /* for (temp = nicklist[ii]; temp; temp = temp->next) */
  } /* for (ii = 0; ii < NICKLIST_MAX; ++ii) */

  notice(n_NickServ, lptr->nick, 
    "-- End of list (%d/%d matches shown) --",
    acnt,
    mcnt);
} /* n_list() */

/*
n_info()
  Display some information about nickname av[1]
*/

static void
n_info(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr, *realptr;
  struct Luser *userptr;
  int online,
      isadmin;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002INFO <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "INFO");
    return;
  }

  if (!(realptr = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  if (!(nptr = GetMaster(realptr)))
    return;

  RecordCommand("%s: %s!%s@%s INFO %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    realptr->nick);

  isadmin = IsValidAdmin(lptr);

  if (((nptr->flags & NS_PRIVATE) || (nptr->flags & NS_FORBID))
  #ifdef EMPOWERADMINS
    && !isadmin)
  #else
    )
  #endif
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] is private",
      realptr->nick);
    RecordCommand("%s: %s!%s@%s failed INFO %s",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      realptr->nick);
    return;
  }

  online = 0;
  if ((userptr = FindClient(realptr->nick)))
    if (realptr->flags & NS_IDENTIFIED)
      online = 1;

  notice(n_NickServ, lptr->nick,
    "           Nickname: %s %s",
    realptr->nick,
    online ? "<< ONLINE >>" : "");

  notice(n_NickServ, lptr->nick,
    "         Registered: %s ago",
    timeago(realptr->created, 1));

  if (realptr->lastseen && !online)
    notice(n_NickServ, lptr->nick,
      "          Last Seen: %s ago",
      timeago(realptr->lastseen, 1));

  if (!(nptr->flags & NS_HIDEALL) || isadmin)
  {
    char buf[MAXLINE];

    if (LastSeenInfo)
    {
      if (!(nptr->flags & NS_HIDEADDR) || isadmin)
        if (!online && realptr->lastu && realptr->lasth)
          notice(n_NickServ, lptr->nick,
            "  Last Seen Address: %s@%s",
            realptr->lastu,
            realptr->lasth);

      if (!(nptr->flags & NS_HIDEQUIT) || isadmin)
        if (realptr->lastqmsg)
          notice(n_NickServ, lptr->nick,
            " Last Seen Quit Msg: %s",
            realptr->lastqmsg);
    } /* if (LastSeenInfo) */

    if (!(nptr->flags & NS_HIDEEMAIL) || isadmin)
      if (realptr->email)
        notice(n_NickServ, lptr->nick,
          "      Email Address: %s", realptr->email);

    if (!(nptr->flags & NS_HIDEURL) || isadmin)
      if (realptr->url)
        notice(n_NickServ, lptr->nick,
          "                URL: %s", realptr->url);

    if (nptr->UIN)
      notice(n_NickServ, lptr->nick,
          "                UIN: %s", nptr->UIN);

    if (nptr->gsm)
      notice(n_NickServ, lptr->nick,
          "                GSM: %s", nptr->gsm);

    if (nptr->phone)
      notice(n_NickServ, lptr->nick,
          "              Phone: %s", nptr->phone);

    buf[0] = '\0';
    if (AllowKillProtection)
      if (nptr->flags & NS_PROTECTED)
        strcat(buf, "Kill Protection, ");
    if (nptr->flags & NS_OPERATOR)
      strcat(buf, "IRC Operator, ");
    if (nptr->flags & NS_AUTOMASK)
      strcat(buf, "AutoMask, ");
    if (nptr->flags & NS_PRIVATE)
      strcat(buf, "Private, ");
    if (nptr->flags & NS_FORBID)
      strcat(buf, "Forbidden, ");
    if (nptr->flags & NS_SECURE)
      strcat(buf, "Secure, ");
    if (nptr->flags & NS_UNSECURE)
      strcat(buf, "UnSecure, ");
    if (nptr->flags & NS_HIDEALL)
      strcat(buf, "Hidden, ");
    if (nptr->flags & NS_MEMOS)
      strcat(buf, "AllowMemos, ");
    if (nptr->flags & NS_MEMONOTIFY)
      strcat(buf, "MemoNotify, ");
    if (nptr->flags & NS_MEMOSIGNON)
      strcat(buf, "MemoSignon, ");
    if (nptr->flags & NS_BGHELP)
      strcat(buf, "BGHelp, ");
    if (nptr->flags & NS_SPOOF)
      strcat(buf, "SpoofAuth, ");

    if (*buf)
    {
      buf[strlen(buf) - 2] = '\0';
      notice(n_NickServ, lptr->nick,
        "   Nickname Options: %s",
        buf);
    }
  }

  if (isadmin || 
    (lptr == FindClient(realptr->nick) && realptr->flags & NS_IDENTIFIED))
  {
    int cnt;

    /*
     * Show administrators all nicknames in this user's
     * link, and also all channels this user has registered
     */

	/* ...and users inquiring about their nick too
	 */
  #ifdef LINKED_NICKNAMES

    if (nptr->master || nptr->nextlink)
    {
      struct NickInfo *tmp;

      cnt = 0;

      notice(n_NickServ, lptr->nick,
        "   Linked Nicknames:");

      for (tmp = nptr->master ? nptr->master : nptr, myi = 0; tmp && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
      {
        notice(n_NickServ, lptr->nick,
          "                     %d) %s",
          ++cnt,
          tmp->nick);
      }
    }

  #endif /* LINKED_NICKNAMES */

  #ifdef CHANNELSERVICES

    /*
     * ChanServDB makes all founder entries the nickname of
     * the LINK master - so make sure we test the founder
     * nickname against the link master's
     */
    if (nptr->FounderChannels)
    {
      struct aChannelPtr *tmpchan;

      notice(n_NickServ, lptr->nick,
        "Channels Registered:");

      cnt = 0;
      for (tmpchan = nptr->FounderChannels; tmpchan; tmpchan = tmpchan->next)
      {
        notice(n_NickServ, lptr->nick,
          "                     %d) %s",
          ++cnt,
          tmpchan->cptr->name);
      }
    }

    if (nptr->AccessChannels)
    {
      struct AccessChannel *acptr;

      notice(n_NickServ, lptr->nick,
        "Access Channels:");

      cnt = 0;
      for (acptr = nptr->AccessChannels; acptr; acptr = acptr->next)
      {
        notice(n_NickServ, lptr->nick,
          "                     %d) (%-3d) %s",
          ++cnt,
          acptr->accessptr->level,
          acptr->cptr->name);
      }
    }

  #endif /* CHANNELSERVICES */

  } /* if (isadmin) */
} /* n_info() */

#ifdef LINKED_NICKNAMES

/*
n_link()
 Link lptr's nickname to av[1]. Make sure lptr supplied the correct
password for the nickname av[1], which will be stored in av[2]
*/

static void
n_link(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *target, /* target nickname */
                  *nptr; /* lptr's nick structure */
  int badlink,
      ret;

  if (ac < 3)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002LINK <nickname> <password>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, 
      n_NickServ,
      "LINK");
    return;
  }

  if (!(target = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  badlink = 0;

  if (!(nptr = FindNick(lptr->nick)))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      lptr->nick);
    badlink = 1;
  }
  else if (target->flags & NS_FORBID)
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] is forbidden",
      target->nick);
    badlink = 1;
  }
  else if (!irccmp(target->nick, lptr->nick))
  {
    notice(n_NickServ, lptr->nick,
      "You cannot link to your current nickname");
    badlink = 1;
  }
  else if (!pwmatch(target->password, av[2]))
  {
    notice(n_NickServ, lptr->nick,
      ERR_BAD_PASS);
    badlink = 1;
  }


  if (badlink)
  {
    RecordCommand("%s: %s!%s@%s failed LINK %s",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      target->nick);
    return;
  }

  RecordCommand("%s: %s!%s@%s LINK %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    target->nick);

  if ((ret = InsertLink(target, nptr)) > 0)
  {
    struct NickHost *htmp;

    /*
     * Successful link - delete lptr's (nptr's) access list,
     * also copy nptr's memo list to target's list
     */
    while (nptr->hosts)
    {
      htmp = nptr->hosts->next;
      MyFree(nptr->hosts->hostmask);
      MyFree(nptr->hosts);
      nptr->hosts = htmp;
    }

  #ifdef MEMOSERVICES
    MoveMemos(nptr, target);
  #endif /* MEMOSERVICES */

    notice(n_NickServ, lptr->nick,
      "Your nickname is now linked to [\002%s\002]",
      target->nick);
  }
  else if (ret == 0)
  {
    notice(n_NickServ, lptr->nick,
      "Link failed");
    putlog(LOG1, "%s: n_link: InsertLink() failed for %s (hub: %s)",
      n_NickServ,
      lptr->nick,
      target->nick);
  }
  else if (ret == (-1))
  {
    notice(n_NickServ, lptr->nick,
      "Link failed: circular link detected");
  }
  else if (ret == (-2))
  {
    notice(n_NickServ, lptr->nick,
      "Number of links cannot exceed %d",
      MaxLinks);
  }
} /* n_link() */

/*
n_unlink()
 Unlink lptr->nick from its current nick link. Make a copy of it's
master's access list for it.
 If av[1]/av[2] supplied, unlink nickname av[1] from its nick link
(making sure password av[2] matches up)
*/

static void
n_unlink(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;
  int unlink;

  if (ac == 2)
  {
    /*
     * They specified a nickname to unlink, but no password
     */
    notice(n_NickServ, lptr->nick,
      "Syntax: \002UNLINK [nickname [password]]\002");
    notice(n_NickServ, lptr->nick, ERR_MORE_INFO, n_NickServ, "UNLINK");
    return;
  }

  if (ac < 3)
    nptr = FindNick(lptr->nick);
  else
    nptr = FindNick(av[1]);

  if (!nptr)
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      (ac < 3) ? lptr->nick : av[1]);
    return;
  }

  if (ac >= 3)
  {
    if (!pwmatch(nptr->password, av[2]))
    {
      notice(n_NickServ, lptr->nick,
        ERR_BAD_PASS);
      RecordCommand("%s: %s!%s@%s failed UNLINK %s",
        n_NickServ, lptr->nick, lptr->username, lptr->hostname,
        nptr->nick);
      return;
    }
  }

  RecordCommand("%s: %s!%s@%s UNLINK %s",
    n_NickServ, lptr->nick, lptr->username, lptr->hostname, nptr->nick);

  unlink = DeleteLink(nptr, 1);

  if (unlink > 0)
  {
    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] has been unlinked",
      nptr->nick);
  }
  else if (unlink == 0)
  {
    notice(n_NickServ, lptr->nick,
      "DeleteLink() reported errors, unlink failed");
    putlog(LOG1, "%s: UNLINK failed: DeleteLink() contained errors",
      n_NickServ);
  }
  else if (unlink == -1)
  {
    notice(n_NickServ, lptr->nick, ERR_NOT_LINKED, nptr->nick);
  }
} /* n_unlink() */

#endif /* LINKED_NICKNAMES */

#ifdef EMPOWERADMINS

/*
n_forbid()
  Create forbidden nickname av[1] (assume permission to exec
this command has already been checked)
*/

static void
n_forbid(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002FORBID <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "FORBID");
    return;
  }

  RecordCommand("%s: %s!%s@%s FORBID [%s]",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  o_Wallops("FORBID from %s for nick [%s]",
    lptr->nick,
    av[1]);

  if (!(nptr = FindNick(av[1])))
  {
    /* nick is not registered - do so now; add nick to nick table */

    nptr = MakeNick();
    nptr->nick = MyStrdup(av[1]);
    nptr->created = current_ts;
    nptr->flags |= NS_FORBID;
    AddNick(nptr);
  }
  else
  {
    /* the nickname is already registered */

    if (nptr->flags & NS_FORBID)
    {
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] is already forbidden",
        av[1]);
      return;
    }

    nptr->flags |= NS_FORBID;
    nptr->flags &= ~NS_IDENTIFIED;
  }

  notice(n_NickServ, lptr->nick,
    "The nickname [\002%s\002] is now forbidden",
    av[1]);

  /*
   * Check if av[1] is currently online, if so, give warning
   */
  CheckNick(av[1]);
} /* n_forbid() */

/*
 * n_unforbid()
 * Remove forbid flag on nickname av[1]
 * (assume permission to exec this command has already been checked)
 * -kre
 */
static void
n_unforbid(struct Luser *lptr, int ac, char **av)
{
  struct NickInfo *nptr;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002UNFORBID <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO, n_NickServ, "UNFORBID");
    return;
  }

  RecordCommand("%s: %s!%s@%s UNFORBID [%s]",
    n_NickServ, lptr->nick, lptr->username, lptr->hostname, av[1]);

  o_Wallops("UNFORBID from %s for nick [%s]", lptr->nick, av[1]);

  if (!(nptr = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick, ERR_NOT_REGGED, av[1]);
    return;
  }

  if (!nptr->password)
  {
    /* It is made from AddNick() in forbid() code - it is empty, so it is
     * safe to delete it -kre */
    DeleteNick(nptr);

    notice(n_NickServ, lptr->nick,
      "The nickname [\002%s\002] has been dropped", av[1]);
  }
  else
  {
    nptr->flags &= ~NS_FORBID;
  
    notice(n_NickServ, lptr->nick,
      "Nickname [\002%s\002] is now unforbidden", av[1]);

    /* Check if av[1] is currently online and tell it to identify -kre */
    CheckNick(av[1]);
  }
} /* n_unforbid() */

/*
n_setpass()
  Change the password for nickname av[1] to av[2]
*/

static void
n_setpass(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 4)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SETPASS <nickname> <password> <securepassword>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SETPASS");
    return;
  }

  RecordCommand("%s: %s!%s@%s SETPASS [%s] ...",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  o_Wallops("SETPASS from %s!%s@%s for nick [%s]",
    lptr->nick, lptr->username, lptr->hostname,
    av[1] );

  if (!(nptr = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick, ERR_NOT_REGGED,
      av[1]);
    return;
  }

  if (!ChangePass(nptr, av[2], av[3]))
  {
    notice(n_NickServ, lptr->nick,
      "Password change failed");
    RecordCommand("%s: Failed to change passwords for [%s] to [%s] and [%s] (SETPASS)",
      n_NickServ,
      av[1],
      av[2], av[3]);
    return;
  }

  notice(n_NickServ, lptr->nick,
    "Passwords for [\002%s\002] has been changed to [\002%s\002] and [\002%s\002]",
    av[1],
    av[2], av[3]);
} /* n_setpass() */

/*
n_noexpire()
 Mark nickname av[1] never to expire
*/

static void
n_noexpire(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002NOEXPIRE <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "NOEXPIRE");
    return;
  }

  RecordCommand("%s: %s!%s@%s NOEXPIRE [%s]",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  if (!(nptr = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  nptr->flags |= NS_OPERATOR;

  notice(n_NickServ, lptr->nick,
    "The nickname [\002%s\002] will never expire",
    nptr->nick);
} /* n_noexpire() */

#ifdef LINKED_NICKNAMES

/*
n_showlink()
 Show nicknames linked to av[1]
*/

static void
n_showlink(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr, *tmp, *start;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002SHOWLINK <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "SHOWLINK");
    return;
  }

  RecordCommand("%s: %s!%s@%s SHOWLINK [%s]",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  if (!(nptr = GetLink(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  if (!nptr->master && !nptr->nextlink)
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_LINKED,
      nptr->nick);
    return;
  }

  notice(n_NickServ, lptr->nick,
    "-- Listing Nickname Link for [\002%s\002] --",
    nptr->nick);

  notice(n_NickServ, lptr->nick,
    "\002%s\002 << MASTER >>",
    nptr->master ? nptr->master->nick : nptr->nick);

  if (nptr->master)
    start = nptr->master->nextlink;
  else
    start = nptr->nextlink;

  for (tmp = start, myi = 0; tmp && myi <= MaxLinks; tmp = tmp->nextlink, myi++)
  {
    notice(n_NickServ, lptr->nick,
      "  %s",
      tmp->nick);

  }

  notice(n_NickServ, lptr->nick,
    "-- End of list (%d nicknames in link) --",
    nptr->master ? nptr->master->numlinks : nptr->numlinks);
} /* n_showlink() */

/*
n_droplink()
 Drop every nickname in the link av[1] is in.
*/

static void
n_droplink(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr, *tmp;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002DROPLINK <nickname>\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "DROPLINK");
    return;
  }

  RecordCommand("%s: %s!%s@%s DROPLINK [%s]",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  if (!(nptr = GetLink(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  if (!nptr->master && !nptr->nextlink)
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_LINKED,
      nptr->nick);
    return;
  }

  while (nptr)
  {
    tmp = nptr->nextlink;
    DeleteNick(nptr);
    nptr = tmp;
  }

  notice(n_NickServ, lptr->nick,
    "Link containing [\002%s\002] has been dropped",
    av[1]);
} /* n_droplink() */

#endif /* LINKED_NICKNAMES */

/*
n_collide()
 Perform various tasks on the current nick collide list

Options:
 -list   - list nicknames to be collided
 -halt   - stop nick collision for targets
 -set    - set nick collision for target
 -setnow - set nick collision for target immediately
*/

static void
n_collide(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;
  int ii, cnt;
  int list, /* 1 if -list given */
      halt, /* 1 if -halt given */
      set, /* 1 if -set given */
      setnow; /* 1 if -setnow given */
  char *target;
  char argbuf[MAXLINE];

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002COLLIDE <pattern> [options]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "COLLIDE");
    return;
  }

  target = NULL;
  list = halt = set = setnow = 0;

  for (cnt = 1; cnt < ac; cnt++)
  {
    if (!ircncmp(av[cnt], "-list", strlen(av[cnt])))
      list = 1;
    else if (!ircncmp(av[cnt], "-halt", strlen(av[cnt])))
      halt = 1;
    else if (!ircncmp(av[cnt], "-set", strlen(av[cnt])))
      set = 1;
    else if (!ircncmp(av[cnt], "-setnow", strlen(av[cnt])))
      setnow = 1;
    else
    {
      if (!target)
        target = av[cnt];
    }
  }

  if (!target)
  {
    notice(n_NickServ, lptr->nick,
      "No pattern specified");
    notice(n_NickServ, lptr->nick,
      "Syntax: \002COLLIDE <pattern> [options]\002");
    return;
  }

  if (ac < 3)
  {
    /*
     * They didn't give any options - default to -set
     */
    set = 1;
  }

  ircsprintf(argbuf, "[%s] ", target);

  if (list)
    strcat(argbuf, "-list ");

  if (halt)
    strcat(argbuf, "-halt ");

  if (set)
    strcat(argbuf, "-set ");

  if (setnow)
    strcat(argbuf, "-setnow ");

  RecordCommand("%s: %s!%s@%s COLLIDE %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    argbuf);

  if (set || setnow)
  {
    nptr = FindNick(target);

    if (!nptr)
    {
      notice(n_NickServ, lptr->nick,
        ERR_NOT_REGGED,
        target);
      return;
    }

    if (set)
    {
      nptr->flags |= NS_COLLIDE;
      nptr->collide_ts = current_ts + 60;
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] has been marked for collision",
        nptr->nick);
    }
    else /* setnow */
    {
      collide(nptr->nick);
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] has been collided",
        nptr->nick);
    }

    return;
  } /* if (set || setnow) */

  if (list)
    notice(n_NickServ, lptr->nick,
      "-- Listing collisions matching [\002%s\002] --",
      target);

  cnt = 0;
  for (ii = 0; ii < NICKLIST_MAX; ++ii)
  {
    for (nptr = nicklist[ii]; nptr; nptr = nptr->next)
    {
      if (nptr->flags & NS_COLLIDE)
      {
        if (match(target, nptr->nick))
        {
          cnt++;
          if (halt)
            nptr->flags &= ~NS_COLLIDE;

          if (list)
          {
            notice(n_NickServ, lptr->nick,
              "%-15s",
              nptr->nick);
          }
        }
      } /* if (nptr->flags & NS_COLLIDE) */
    }
  }

  if (list)
  {
    notice(n_NickServ, lptr->nick,
      "-- End of list (%d match%s found) --",
      cnt,
      (cnt == 1) ? "" : "es");
  }
  else
  {
    notice(n_NickServ, lptr->nick,
      "%d match%s found",
      cnt,
      (cnt == 1) ? "" : "es");
  }
} /* n_collide() */

/*
n_flag()
 Gives a registered nickname various flags which will limit
their privileges

Available flags:
-noregister     Prevents them from registering a channel
-nochanops      Will not op them on a channel if they're an AOP
-nospoof		Forbid from using HostServ's SPOOF
-clear          Clear all current flags
*/

static void
n_flag(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr, *realptr;
  char buf[MAXLINE],
       pstr[MAXLINE];
  int ii;

  if (ac < 2)
  {
    notice(n_NickServ, lptr->nick,
      "Syntax: \002FLAG <nickname> [options]\002");
    notice(n_NickServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "FLAG");
    return;
  }

  if (!(realptr = FindNick(av[1])))
  {
    notice(n_NickServ, lptr->nick,
      ERR_NOT_REGGED,
      av[1]);
    return;
  }

  if (!(nptr = GetMaster(realptr)))
    return;

  buf[0] = '\0';
  pstr[0] = '\0';

  if (ac < 3)
  {
    RecordCommand("%s: %s!%s@%s FLAG [%s]",
      n_NickServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      realptr->nick);

    /*
     * They gave no options, just display a list of
     * nptr's current flags
     */
    if (nptr->flags & NS_NOREGISTER)
      strcat(buf, "NoRegister, ");
    if (nptr->flags & NS_NOCHANOPS)
      strcat(buf, "NoChannelOps, ");
    if (nptr->flags & NS_NOSPOOF)
      strcat(buf, "NoSpoof, ");

    if (*buf)
    {
      buf[strlen(buf) - 2] = '\0';
      notice(n_NickServ, lptr->nick,
        "The flags for [\002%s\002] are: %s",
        realptr->nick,
        buf);
    }
    else
    {
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] has no current flags",
        realptr->nick);
    }

    return;
  }

  /*
   * Parse the options they gave
   */
  for (ii = 2; ii < ac; ++ii)
  {
    if (!ircncmp(av[ii], "-noregister", strlen(av[ii])))
    {
      nptr->flags |= NS_NOREGISTER;
      strcat(buf, "NoRegister, ");
      strcat(pstr, "-noregister ");
    }
    else if (!ircncmp(av[ii], "-nochanops", strlen(av[ii])))
    {
      nptr->flags |= NS_NOCHANOPS;
      strcat(buf, "NoChannelOps, ");
      strcat(pstr, "-nochanops ");
    }
    else if (!ircncmp(av[ii], "-nospoof", strlen(av[ii])))
    {
      nptr->flags |= NS_NOSPOOF;
      strcat(buf, "NoSpoof, ");
      strcat(pstr, "-nospoof ");
    }
    else if (!ircncmp(av[ii], "-clear", strlen(av[ii])))
    {
      nptr->flags &= ~(NS_NOREGISTER | NS_NOCHANOPS | NS_NOSPOOF);
      notice(n_NickServ, lptr->nick,
        "The flags for [\002%s\002] have been cleared",
        realptr->nick);
      buf[0] = '\0';
      strcat(pstr, "-clear ");

      break;
    }
    else
    {
      notice(n_NickServ, lptr->nick,
        "Invalid flag: %s",
        av[ii]);
    }
  }

  if (*buf)
  {
    buf[strlen(buf) - 2] = '\0';
    notice(n_NickServ, lptr->nick,
      "The flags (%s) have been added to [\002%s\002]",
      buf,
      realptr->nick);
  }

  RecordCommand("%s: %s!%s@%s FLAG [%s] %s",
    n_NickServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    realptr->nick,
    pstr);
} /* n_flag() */

#endif /* EMPOWERADMINS */

/*
 */
static void n_fixts(struct Luser *lptr, int ac, char **av)
{
  int tsdelta = 0;
  time_t now = 0;
  char dMsg[] = "Detected nickname \002%s\002 with TS %d "
                "below TS_MAX_DELTA %d";
  struct Luser *ouser = NULL;

  if (ac < 2)
    tsdelta = MaxTSDelta;
  else
    tsdelta = atoi(av[1]);

  now = current_ts;

  /* Be paranoid */
  if (tsdelta <= 0)
  {
    notice(n_ChanServ, lptr->nick,
        "Wrong TS_MAX_DELTA specified, using default of 8w");
    tsdelta = 4838400; /* 8 weeks */
  }

  for (ouser = ClientList; ouser; ouser = ouser->next)
  {
    if ((now - ouser->since) >= tsdelta)
    {
      SendUmode(OPERUMODE_Y, dMsg, ouser->nick, ouser->since, tsdelta);
      notice(n_NickServ, lptr->nick, dMsg, ouser->nick, ouser->since,
          tsdelta);
      putlog(LOG1, "%s: Bogus TS nickname: [%s] (TS=%d)", 
          n_ChanServ, ouser->nick, ouser->since);

      collide(ouser->nick);
      notice(n_NickServ, lptr->nick,
        "The nickname [\002%s\002] has been collided",
        ouser->nick);
    }
  }
}

#endif /* NICKSERVICES */
